<?php
/**
* LICENCIA
*
* Este programa se propociona "tal cual", sin garantía de ningún tipo más allá del soporte
* pactado a la hora de adquirir el programa.
*
* En ningún caso los autores o titulares del copyright serán responsables de ninguna
* reclamación, daños u otras responsabilidades, ya sea en un litigio, agravio o de otro
* modo, que surja de o en conexión con el programa o el uso u otro tipo de acciones
* realizadas con el programa.
*
* Este programa no puede modificarse ni distribuirse sin el consentimiento expreso del autor.
*
*    @author    Carlos Fillol Sendra <festeweb@festeweb.com>
*    @copyright 2014 Fes-te web! - www.festeweb.com
*    @license   http://www.festeweb.com/static/licenses/fs2ps_1.1.0.txt
*/

include_once(dirname(__FILE__).'/Fs2psDto2RowMatcher.php');
include_once(dirname(__FILE__).'/Fs2psException.php');
include_once(dirname(__FILE__).'/Fs2psTools.php');

class Fs2psDto2RowRefMatcher extends Fs2psDto2RowMatcher
{

  public function __construct($entity, $table, $row_id_field, $dto_id_fields, $persist = false)
	{
	    parent::__construct($entity, $table, $row_id_field, $dto_id_fields, $persist);
	}

	public function _rowIdFromDtoId($dto_id)
	{
		$dto_id_sql = Fs2psTools::dbEscape($this->dtoIdToStr($dto_id));
		return (int)Fs2psTools::dbValue('
			SELECT '.$this->row_id_field.'
			FROM `@DB_'.$this->table.'`
			WHERE `reference` = \''.$dto_id_sql.'\'
		');
	}

}

class Fs2psDto2RowPRefMatcher extends Fs2psDto2RowMatcher
{
    
    public function __construct($entity, $table, $row_id_field, $dto_id_fields, $persist = true)
    {
        parent::__construct($entity, $table, $row_id_field, $dto_id_fields, $persist);
    }
    
    /**
     * Sólo debe usarse esta función en caso de que dto tenga suficiente información
     * como en el momento de la creación/actualización.
     * 
     * Si existe un producto con matching por pref -> matching (si hay varios se cogerá el primero)
     * En caso contrario se intenta el matching básico por ref (cod FS) 
     */
    public function rowIdFromDto($dto)
    {
        $dto_id = $this->dtoId($dto);
        $row_id = $this->rowIdFromDtoIdForcedMatches($dto_id);
        if ($row_id) return $row_id;
        
        if (!empty($dto['pref']))
        {
            $dto_pref_sql = Fs2psTools::dbEscape($dto['pref']);
            $row_id = (int)Fs2psTools::dbValue('
    			SELECT '.$this->row_id_field.'
    			FROM `@DB_'.$this->table.'`
    			WHERE `reference` = \''.$dto_pref_sql.'\'
    		');
            if ($row_id) return $row_id;
        }
        
        return $this->_rowIdFromDtoId($dto_id);
    }
    
}

class Fs2psDto2RowCustomerEmailMatcher extends Fs2psDto2RowMatcher
{
    
    public function __construct($entity, $table, $row_id_field, $dto_id_fields, $persist = false)
    {
        parent::__construct($entity, $table, $row_id_field, $dto_id_fields, $persist);
    }
    
    /**
     * Sólo debe usarse esta función en caso de que dto tenga suficiente información
     * como en el momento de la creación/actualización.
     *
     * Si existe un producto con matching por email -> matching (si hay varios se cogerá el primero)
     * En caso contrario se intenta el matching básico por ref (cod FS)
     */
    public function rowIdFromDto($dto)
    {
        if (!empty($dto['email']))
        {
            $dto_pref_sql = Fs2psTools::dbEscape($dto['email']);
            $row_id = (int)Fs2psTools::dbValue('
    			SELECT '.$this->row_id_field.'
    			FROM `@DB_'.$this->table.'`
    			WHERE `email` = \''.$dto_pref_sql.'\'
                ORDER BY id_customer
                LIMIT 1
    		');
            if ($row_id) return $row_id;
        }
        
        return $this->rowIdFromDtoId($this->dtoId($dto));
    }
    
}

class Fs2psDto2RowAddressNifMatcher extends Fs2psDto2RowMatcher
{
    protected $task;
    
    public function __construct($entity, $table, $row_id_field, $dto_id_fields, $persist=false, $task=null)
    {
        $this->task = $task;
        parent::__construct($entity, $table, $row_id_field, $dto_id_fields, $persist);
    }
    
    /**
     * Sólo debe usarse esta función en caso de que dto tenga suficiente información
     * como en el momento de la creación/actualización.
     *
     * Si existe un producto con matching por email -> matching (si hay varios se cogerá el primero)
     * En caso contrario se intenta el matching básico por ref (cod FS)
     */
    public function rowIdFromDto($dto)
    {
        if ($this->task && !empty($dto['nif']) && !empty($dto['customer']))
        {
            $id_customer = $this->task->getUpdater('customers')->matcher->rowIdFromDtoId($dto['customer']);
            if ($id_customer) {
                $dto_pref_sql = Fs2psTools::dbEscape($dto['nif']);
                $row_id = (int)Fs2psTools::dbValue('
        			SELECT '.$this->row_id_field.'
        			FROM `@DB_'.$this->table.'`
        			WHERE `dni` = \''.$dto_pref_sql.'\' and id_customer = '.$id_customer.' 
                    ORDER BY id_address
                    LIMIT 1
        		');
                if ($row_id) return $row_id;
            }
        }
        
        return $this->rowIdFromDtoId($this->dtoId($dto));
    }
    
}