<?php
/**
* LICENCIA
*
* Este programa se propociona "tal cual", sin garantía de ningún tipo más allá del soporte
* pactado a la hora de adquirir el programa.
*
* En ningún caso los autores o titulares del copyright serán responsables de ninguna
* reclamación, daños u otras responsabilidades, ya sea en un litigio, agravio o de otro
* modo, que surja de o en conexión con el programa o el uso u otro tipo de acciones
* realizadas con el programa.
*
* Este programa no puede modificarse ni distribuirse sin el consentimiento expreso del autor.
*
*    @author    Carlos Fillol Sendra <festeweb@festeweb.com>
*    @copyright 2014 Fes-te web! - www.festeweb.com
*    @license   http://www.festeweb.com/static/licenses/fs2ps_1.1.0.txt
*/

include_once(dirname(__FILE__).'/Fs2psTools.php');
include_once(dirname(__FILE__).'/Fs2psExtractor.php');
include_once(dirname(__FILE__).'/Fs2psDto2RowRefMatcher.php');
include_once(dirname(__FILE__).'/Fs2psObjectModels.php');
include_once(dirname(__FILE__).'/Fs2psUpdaters.php');


class Fs2psOrderDependentExtractor extends Fs2psExtractor
{
	protected $download_orders;
	protected $ifvalid;

	protected function reloadCfg() {
		parent::reloadCfg();
		$cfg = $this->task->cfg;
		$this->download_orders = $cfg->get('DOWNLOAD_ORDERS', '');
		$this->ifvalid = strpos($this->download_orders, 'ifvalid') !== false;
	}

	protected function getAfterDateWhereCondition()
	{
		$where = array('o.current_state<>6');

		$after_str = '1970-01-01 01:00:00';
		if (!empty($this->task->cmd['after']))
			$after_str =  Fs2psTools::date2db(Fs2psTools::dto2date($this->task->cmd['after']));
		$where[] = 'o.date_upd>\''.$after_str.'\'';

		if (!empty($this->task->cmd['until']))
		{
			$until_str =  Fs2psTools::date2db(Fs2psTools::dto2date($this->task->cmd['until']));
			$where[] = 'o.date_upd<=\''.$until_str.'\'';
		}

		if ($this->ifvalid) $where[] = 'o.valid=1';

		return join(" and ", $where);
	}
}

class Fs2psCustomerExtractor extends Fs2psOrderDependentExtractor
{
	
	protected $download_customers;
	protected $noorder;
	protected $noaddress;
	protected $optional_fields = array('address2', 'extra1', 'other');
	
	protected function reloadCfg() {
		parent::reloadCfg();
		$cfg = $this->task->cfg;
		$this->download_customers = $cfg->get('DOWNLOAD_CUSTOMERS', '');
		$this->noaddress = strpos($this->download_customers, 'noaddress') !== false;
		$this->noorder = $this->noaddress || strpos($this->download_customers, 'noorder') !== false;
	}
	
	protected function getAfterDateWhereCondition()
	{
		if (empty($this->task->cmd['after']))
			$dto_after = '1970-01-01 01:00:00';
		else
			$dto_after = $this->task->cmd['after'];

		$after =  Fs2psTools::dto2date($dto_after);
		$after_str = Fs2psTools::date2db($after);
		
		$download_orders_where = '';
		if (!$this->noorder) {
			$download_orders_where .= ' and o.current_state<>6 ';
		   	if ($this->ifvalid) $download_orders_where .= ' and o.valid=1 ';
		}
		
		return '((c.date_upd>\''.$after_str.'\' or a.date_upd>\''.$after_str.'\' or o.date_upd>\''.$after_str.'\')'.$download_orders_where.')';
	}

	protected function buildSql()
	{

		// XXX cfillol: Cuidado con los selects con columnas que no estan en el group by
		// su soporte puede estar deshabilitado en algunos MySQLs y el orden puede ser arbitrario.
		// https://dev.mysql.com/doc/refman/5.0/en/group-by-handling.html
		// http://renormalize.org/post/24114561260/mysql-curious-features-non-deterministic-group-by
		
		// XXX cfillol: ¿Qué pasa si hay varias direcciones del mismo cliente con el mismo nif?
		// Puede que no se cargue la correcta, sobretodo si se activa 'noorder'
		
		return '
			SELECT
			    c.id_customer, a.id_address,
			    greatest(c.date_add, coalesce(a.date_add, 0)) as date_add, 
			    greatest(c.date_upd, coalesce(a.date_upd, 0), coalesce(o.date_upd, 0)) as date_upd,
                c.email,
			    IFNULL(a.firstname, c.firstname) as firstname,
                IFNULL(a.lastname, c.lastname) as lastname,
			    a.company,
				c.company as extra1,
                a.other as other,
			    a.address1, a.address2, a.postcode,
			    a.city, a.phone, a.phone_mobile,
			    a.vat_number, a.dni,
			    st.name as state, co.name as country
			FROM
			    @DB_customer c
			    inner join (
					SELECT 
					    id_customer, nif,
					    CONVERT(substring_index(ordadd,\'|\',1), UNSIGNED INTEGER) as id_order, 
					    CONVERT(substring_index(substring_index(ordadd,\'|\',2), \'|\',-1), UNSIGNED INTEGER) as id_address
					FROM (
					    SELECT
					        c.id_customer,
					        max(concat(LPAD(ifnull(o.id_order, 0),10,\'0\'), \'|\', LPAD(a.id_address,10,\'0\'))) as ordadd,
					        REPLACE(REPLACE(upper(ifnull(a.dni, vat_number)),\' \',\'\'),\'-\',\'\') as nif
					    FROM
					        @DB_customer c
					        '.($this->noaddress? 'left' : 'inner').' join @DB_address a on a.id_customer=c.id_customer and a.deleted=0 -- and a.active=1 
					        '.($this->noorder? 'left' : 'inner').' join @DB_orders o on o.id_address_invoice=a.id_address
					    WHERE
					       '.$this->getAfterDateWhereCondition().'
					       		-- XXX cfillol: Puede que hayan direcciones modificadas enlazadas en pedidos válidos.
					       		-- Entonces deberemos forzar descarga con DOWNLOAD_CUSTOMERS = False. 
					    group by c.id_customer, nif    
					) ca
			    ) as ai on ai.id_customer = c.id_customer
			    left join @DB_address a on a.id_address = ai.id_address
			    left join @DB_orders o on o.id_order = ai.id_order
			    left join @DB_state st on st.id_state=a.id_state
			    left join @DB_country_lang co
			        on co.id_country=a.id_country and co.id_lang=1
			WHERE
			    '.$this->getAfterDateWhereCondition().'
			ORDER BY date_upd
		';
	}

	protected function row2dto($row)
	{
		$dto = array(
			'user' => intval($row['id_customer']),
			'created' => $row['date_add'],
			'updated' => $row['date_upd'],
			'company' => $row['company'],
			'nif' => $row['dni'],
			'vat_number' => $row['vat_number'],
			'firstname' => $row['firstname'],
			'lastname' => $row['lastname'],
			'email' => $row['email'],
			'address' => $row['address1'],
			'postcode' => $row['postcode'],
			'city' => $row['city'],
			'state' => $row['state'],
			'country' => $row['country'],
			'phone' => $row['phone'],
			'mobile' => $row['phone_mobile'],
		);
		
		foreach ($this->optional_fields as $of) {
		    if (!empty($row[$of])) $dto[$of] = $row[$of];
		}
		
		return $dto;
	}

}


class Fs2psOrderExtractor extends Fs2psOrderDependentExtractor
{
    protected $orders_have_payment_fee; // @Deprecated
	protected $payment_fee_field;
	protected $payment_fee_rate_field;
	
	protected $optional_fields = array('address2', 'saddress2', 'other', 'sother');
	
	protected function reloadCfg() {
		parent::reloadCfg();
		$cfg = $this->task->cfg;
		
		$this->orders_have_payment_fee = $cfg->get('ORDERS_HAVE_PAYMENT_FEE', false); // @Deprecated
		$this->payment_fees = $cfg->get('ORDERS_PAYMENT_FEES');
	}
	
	protected function paymentFeeSelect() {
	    if ($this->orders_have_payment_fee && empty($this->payment_fees)) {
	        // Old deprecated behaviour
		    return 'o.payment_fee, o.payment_fee_rate,';
        }
        
        $select = array();
        foreach ($this->payment_fees as $pf_payment => $pf_cfg) {
            if (!empty($pf_cfg['field'])) $select[] = '`'.Fs2psTools::dbFieldName($pf_cfg['field']).'` as fs2ps_'.$pf_payment.'_payfee';
            if (!empty($pf_cfg['tax_field'])) $select[] = '`'.Fs2psTools::dbFieldName($pf_cfg['tax_field']).'` as fs2ps_'.$pf_payment.'_payfeerate';
        }
        return join($select, ', ').($select? ', ' : '');
	}
	
	protected function buildSql()
	{
		
		$id_default_lang = Configuration::get('PS_LANG_DEFAULT');
		
		return '
			select 
				o.id_order, o.reference, 
				o.invoice_number, 
				o.date_add, o.date_upd,
				o.total_products,
				o.total_paid_tax_incl,
				o.carrier_tax_rate,
				o.total_shipping_tax_incl, o.total_shipping_tax_excl,
				o.total_discounts_tax_excl,
				o.valid,
				c.id_customer, 
                c.email,
				
			    IFNULL(a.firstname, c.firstname) as firstname,
                IFNULL(a.lastname, c.lastname) as lastname,
				a.company, a.vat_number, a.dni,
				
			    a.address1, a.address2, a.postcode, 
			    a.city, a.phone, a.phone_mobile,
				st.name as state, co.name as country,
                a.other as other,
				
				sa.address1 as saddress1, sa.address2 as saddress2, sa.postcode as spostcode, 
			    sa.city as scity, sa.phone as sphone, sa.phone_mobile as sphone_mobile,
				sst.name as sstate, sco.name as scountry,
                sa.other as sother,
				
				'.$this->paymentFeeSelect().'
				o.module as payment,

                (select min(id_message) from @DB_message msg where msg.id_order=o.id_order and msg.private=0) as id_message
				
			FROM
				@DB_orders o
				inner join @DB_customer c on c.id_customer = o.id_customer  -- cfillol XXX: Evitamos descargar pedidos sin cliente (borrado)
				
			    left join @DB_address a on a.id_address = o.id_address_invoice
				left join @DB_state st on st.id_state=a.id_state
				left join @DB_country_lang co
					on co.id_country=a.id_country and co.id_lang='.$id_default_lang.'
							
				left join @DB_address sa on sa.id_address = o.id_address_delivery
				left join @DB_state sst on sst.id_state=sa.id_state
				left join @DB_country_lang sco
					on sco.id_country=sa.id_country and sco.id_lang='.$id_default_lang.'
							
			WHERE
				'.$this->getAfterDateWhereCondition().'
			ORDER BY o.date_upd, o.id_order
		';
	}
	
	protected function getOrderTotals($row)
	{
		$totals_rows = Fs2psTools::dbSelect('
			select
				t.rate as tax,
				sum(od.total_price_tax_excl) as line_base,
				sum(odt.total_amount) as line_tax,
				sum(od.product_quantity) as quantity
			FROM
				@DB_order_detail od
				left join @DB_order_detail_tax odt on odt.id_order_detail=od.id_order_detail
				left join @DB_tax t on t.id_tax=odt.id_tax
			WHERE od.id_order='.$row['id_order'].'
			GROUP BY t.rate
			ORDER BY t.rate
		');

		$dec = 1000; // decimales de precisión para impuestos
		
		$taxes_by_rate = array();
		$lines_tax_amounts_by_tax = array();
		$lines_bases_by_tax = array();
		$discount_bases_by_tax = array();
		$shipping_tax_amounts_by_tax = array();
		$shipping_bases_by_tax = array();
		
		$carrier_tax_rate = null;
		if ($row['total_shipping_tax_excl']>0) {
    		$carrier_tax_rate = floatval($row['carrier_tax_rate']);
    		$carrier_tax_rate_key = intval($carrier_tax_rate * $dec);
    		$taxes_by_rate[$carrier_tax_rate_key] = $carrier_tax_rate;
    		$lines_tax_amounts_by_tax[$carrier_tax_rate_key] = 0;
    		$lines_bases_by_tax[$carrier_tax_rate_key] = 0;
    		$discount_bases_by_tax[$carrier_tax_rate_key] = 0;
    		$shipping_tax_amounts_by_tax[$carrier_tax_rate_key] = 0;
    		$shipping_bases_by_tax[$carrier_tax_rate_key] = 0;
		}
		
		$total_products_base = floatval($row['total_products']);
		$discount_ratio = $total_products_base>0? ($total_products_base-floatval($row['total_discounts_tax_excl']))/$total_products_base : 1;
		foreach ($totals_rows as $totals_row)
		{
			$rate = floatval($totals_row['tax']);
			$rate_key = intval($rate * $dec);
			$taxes_by_rate[$rate_key] = $rate;
			$lines_bases_by_tax[$rate_key] = floatval($totals_row['line_base']);
			$lines_tax_amounts_by_tax[$rate_key] = floatval($totals_row['line_tax']);
			$discount_bases_by_tax[$rate_key] = Tools::ps_round($lines_bases_by_tax[$rate_key] * (1 - $discount_ratio), 2);
			
			// Nos aseguramos de que estén inicializadas a 0 otras tasas distintas a la del transporte
			$shipping_tax_amounts_by_tax[$rate_key] = 0; 
			$shipping_bases_by_tax[$rate_key] = 0;
		}
		if (!is_null($carrier_tax_rate)) {
    		$shipping_tax_amounts_by_tax[$carrier_tax_rate_key] = floatval($row['total_shipping_tax_incl']) - floatval($row['total_shipping_tax_excl']);
    		$shipping_bases_by_tax[$carrier_tax_rate_key] = floatval($row['total_shipping_tax_excl']);
		}
	
		$taxes = array();
		$lines_tax_amounts = array();
		$lines_bases = array();
		$discount_bases = array();
		$shipping_tax_amounts = array();
		$shipping_bases = array();
		foreach ($taxes_by_rate as $rate_key => $val)
		{
		    $taxes[] = $rate_key/$dec;
			$lines_tax_amounts[] = $lines_tax_amounts_by_tax[$rate_key];
			$lines_bases[] = $lines_bases_by_tax[$rate_key];
			$discount_bases[] = $discount_bases_by_tax[$rate_key];
			$shipping_tax_amounts[] = $shipping_tax_amounts_by_tax[$rate_key];
			$shipping_bases[] = $shipping_bases_by_tax[$rate_key];
		}
	
		// Correcciones contables. ¿Que sucede si hay descuadres?
		$total = floatval($row['total_paid_tax_incl']);
		$total_calc = (
		    (array_sum($lines_bases) + array_sum($shipping_bases)) - array_sum($discount_bases) +
		    (array_sum($lines_tax_amounts) + array_sum($shipping_tax_amounts))
		);
		$error = round($total_calc - $total, 2); // redondeamos xq a veces se generan periodos aunque sean sumas y restas :/
		if ($error>0) {
			// Si el descuadre es positivo, lo descontamos del total.
			// Nota: Esto puede provocar que se pague más IVA del necesario.
			foreach ($lines_bases as $i => $base) {
				if (($base-$discount_bases[$i]) >= $error) {
					$discount_bases[$i] += $error;
					break;
				}
			}
		}
		else if ($error<0) {
			// Descuadres en Prestashop?! Puede que el usuario pagara más de lo debido.
			// No debería suceder a menos que exista un bug en los cálculos en Prestashop (ej: PS 1.6.0.9)
			// o exista un bug en el conector ...
			$total = $total + $error;
		}
	
		return array(
			'taxes' => $taxes,
			'lines_tax_amounts' => $lines_tax_amounts,
			'lines_bases' => $lines_bases,
			'discount_bases' => $discount_bases,
			'shipping_tax_amounts' => $shipping_tax_amounts,
			'shipping_bases' => $shipping_bases,
			'total' => $total
		);
	}
	
	protected function row2dto($row)
	{
		$totals = $this->getOrderTotals($row);
					
		$dto = array(
			'id' => intval($row['id_order']),
			'ref' => $row['reference'],
			'invoice' => $row['invoice_number'],
			//'invoice' => intval($row['id_order_invoice']),
			'created' => $row['date_add'],
			'updated' => $row['date_upd'],
			'taxes' => $totals['taxes'],
			'discount_bases' => $totals['discount_bases'],
			'lines_tax_amounts' => $totals['lines_tax_amounts'],
			'lines_bases' => $totals['lines_bases'],
			'shipping_tax_amounts' => $totals['shipping_tax_amounts'],
			'shipping_bases' => $totals['shipping_bases'],
			'total' => $totals['total'],
				
			'company' => empty($row['company'])? $row['firstname'].' '.$row['lastname'] : $row['company'],
		    'firstname' => $row['firstname'],
		    'lastname' => $row['lastname'],
			'user' => intval($row['id_customer']),
			'nif' => $row['dni'],
			'vat_number' => $row['vat_number'],
				
			'address' => $row['address1'],
			'postcode' => $row['postcode'],
			'city' => $row['city'],
			'state' => $row['state'],
			'country' => $row['country'],
			'phone' => $row['phone'],
			'mobile' => $row['phone_mobile'],
			'email' => $row['email'],
				
			'saddress' => $row['saddress1'],
			'spostcode' => $row['spostcode'],
			'scity' => $row['scity'],
			'sstate' => $row['sstate'],
			'scountry' => $row['scountry'],
			'sphone' => $row['sphone'],
			'smobile' => $row['sphone_mobile'],
				
			'valid' => intval($row['valid'])==1,
			'payment' => $row['payment'],
		);
		
		foreach ($this->optional_fields as $of) {
            if (!empty($row[$of])) $dto[$of] = $row[$of];
		}
		
		$payment = $dto['payment'];
		if (!empty($this->payment_fees[$payment])) {
		    if (isset($row['fs2ps_'.$payment.'_payfee'])) {
    		    $dto['payment_fee'] = floatval($row['fs2ps_'.$payment.'_payfee']);
    		}
    		if (isset($row['fs2ps_'.$payment.'_payfeerate'])) {
    		    $dto['payment_fee_rate'] = floatval($row['fs2ps_'.$payment.'_payfeerate']);
    		}
		} else if ($this->orders_have_payment_fee) { // Deprecated
		    $dto['payment_fee'] = floatval($row['payment_fee']);
		    $dto['payment_fee_rate'] = floatval($row['payment_fee_rate']);
		}
		
		if (!empty($row['id_message'])) {
		    $dto['note'] = Fs2psTools::dbValue('select message from `@DB_message` where id_message='.$row['id_message'].'');
		}
		
		return $dto;
	}
}


class Fs2psOrderLineExtractor extends Fs2psOrderDependentExtractor
{
    
    protected $name_pattern;
    protected $name_pattern_langs;
    protected $name_pattern_attributes_langs;
    protected $name_pattern_attributes_fields;
    
    protected $productDto2RowMatcher;
    protected $combinationDto2RowMatcher;
    
    protected function reloadCfg() {
        parent::reloadCfg();
        $cfg = $this->task->cfg;
        
        $this->name_pattern = $cfg->get('ORDER_LINES_NAME_PATTERN');
        
        $match = null;
        preg_match_all('/(?:name|descrip|attributes)_([a-z]{2,3})/', $this->name_pattern, $match);
        $this->name_pattern_langs = array_unique($match[1]);
        
        preg_match_all('/attributes_([a-z]{2,3})/', $this->name_pattern, $match);
        $this->name_pattern_attributes_fields = array_unique($match[0]);
        $this->name_pattern_attributes_langs = array_unique($match[1]);
    }

	public function __construct($task, $name)
	{
		parent::__construct($task, $name);
		$this->productDto2RowMatcher = Fs2psProductUpdater::createMatcher($task, 'products');
		$this->combinationDto2RowMatcher = Fs2psSizeColourCombinationUpdater::createMatcher($task, 'combinations');
	}
	
	protected function buildSqlMultilang($lcods, $tpl, $sep='')
	{
	    if (empty($lcods)) return '';
	    $text = array();
	    $vars = array('$lcod' => NULL);
	    foreach ($lcods as $lcod) {
	        if ($lcod=='ord') continue;
	        $vars['$lcod'] = $lcod;
	        $text[] = strtr($tpl, $vars);
	    }
	    return implode($sep, $text);;
	}
	
	protected function buildSql()
	{
	    $lcods = $this->name_pattern_langs;
		return '
			select
				o.id_order,
				od.id_order_invoice,
                od.product_id,
				od.product_attribute_id,
				od.product_reference,
				od.product_name,
                o.id_lang as lord,
				pl_ord.name as name_ord,
                pl_ord.description_short as descrip_ord,
           '.$this->buildSqlMultilang($lcods, '
				pl_$lcod.name as name_$lcod,
                pl_$lcod.description_short as descrip_$lcod,
           ').'
				od.product_quantity,
				t.rate,
				od.product_price,
				od.unit_price_tax_incl,
				od.unit_price_tax_excl,
				od.reduction_percent,
				od.total_price_tax_incl,
				odt.unit_amount,
				odt.total_amount
			FROM
				@DB_orders o
				left join @DB_order_detail od on od.id_order=o.id_order
				left join @DB_order_detail_tax odt on odt.id_order_detail=od.id_order_detail
				left join @DB_tax t on t.id_tax=odt.id_tax
				left join @DB_product_lang pl_ord on pl_ord.id_product=od.product_id and pl_ord.id_lang=o.id_lang and pl_ord.id_shop=o.id_shop
           '.$this->buildSqlMultilang($lcods, '
				left join @DB_lang l_$lcod on l_$lcod.iso_code=\'$lcod\'
				left join @DB_product_lang pl_$lcod on pl_$lcod.id_product=od.product_id and pl_$lcod.id_lang=l_$lcod.id_lang and pl_$lcod.id_shop=o.id_shop
           ').'
			WHERE
				'.$this->getAfterDateWhereCondition().'
			ORDER BY
				o.date_upd, o.id_order, od.id_order_detail
		';
		/*		NOTE cfillol: No lo usamos. Cogeremos la ref de attributos de la product_reference
		 		left join @DB_product_attribute pa on pa.id_product_attribute=od.product_attribute_id
				left join @DB_product_attribute_combination pac on pac.id_product_attribute=pa.id_product_attribute
				left join @DB_fs2ps_match am on am.row_id=pac.id_attribute and am.table=\'attribute\'
		 */
	}
	
	
	protected function getAttributesMultilang($lcods, $lord, $product_attribute_id, $sep=' ')
	{
        $attributes_by_lang = array();
        if (!empty($this->name_pattern_attributes_fields)) {
            $rows = Fs2psTools::dbSelect('
    		    select
                  '.$this->buildSqlMultilang($lcods, '
                    al_$lcod.name as attribute_$lcod,
                  ').'
                    al_ord.name as attribute_ord
    		    from @DB_product_attribute_combination pac
    		    left join @DB_attribute a on a.id_attribute=pac.id_attribute
                left join @DB_attribute_lang al_ord on al_ord.id_attribute=a.id_attribute and al_ord.id_lang='.$lord.'
    		  '.$this->buildSqlMultilang($lcods, '
    		    left join @DB_lang l_$lcod on l_$lcod.iso_code=\'$lcod\'
                left join @DB_attribute_lang al_$lcod on al_$lcod.id_attribute=a.id_attribute and al_$lcod.id_lang=l_$lcod.id_lang
              ').'
                where pac.id_product_attribute='.$product_attribute_id.'
    		');
            foreach ($this->name_pattern_attributes_langs as $attribute_lang) {
                $attributes = array();
                foreach ($rows as $row) $attributes[] = $row['attribute_'.$attribute_lang];
                $attributes_by_lang['attributes_'.$attribute_lang] = implode($sep, $attributes);
            }
        }
        return $attributes_by_lang;
	}
	
	protected function getTextMultilang($row)
	{
	    if ($this->name_pattern) {
    	    $lcods = $this->name_pattern_langs;
    	    if (!empty($row['product_attribute_id'])) {
    	        $row = array_merge($row, $this->getAttributesMultilang($lcods, $row['lord'], $row['product_attribute_id']));
    	    } else {
    	        foreach ($this->name_pattern_attributes_fields as $field) {
    	            if (!isset($row[$field])) $row[$field] = '';
    	        }
    	    }
    	    return Fs2psTools::htmlToPlainText(strtr($this->name_pattern, $row));
	    } else {
	        return $row['product_name'];
	    }
	}
	
	protected function row2dto($row)
	{
	    if (intval($row['product_attribute_id'])) {
	        $m = $this->combinationDto2RowMatcher;
	        $dto_id_str = $m->persist? $m->dtoIdStrFromRowId($row['product_attribute_id']): $row['product_reference'];
	        if ($m->persist && !$dto_id_str) $dto_id_str = $row['product_reference']; // Puede que queden por descargar pedidos anteriores al cambio de matcher
	        $dto_id = $m->strToDtoId($dto_id_str);
	    } else {
	        $m = $this->productDto2RowMatcher;
	        $dto_id = $m->persist? $m->dtoIdStrFromRowId($row['product_id']): $row['product_reference'];
	        if ($m->persist && !$dto_id) $dto_id = $row['product_reference']; // Puede que queden por descargar pedidos anteriores al cambio de matcher
	    }
		
		$is_array = is_array($dto_id);
		
		///////
		// OJO!!! Puede haber un descuadre considerable la suma de las bases de las lineas y la base total de la factura.
        // Esto sucede cuando se calculan impuestos sobre cada unidad vendida y no sobre el total de la linea. 
		// Se suele hacer así cuando se vende incluyendo el IVA en el precio. Además, se pagan menos impuestos.
		// Cambiamos cálculos por valores en BD a partir de 2.5.2, 2.7.1, 2.6.2
		
		// product_price es el precio unitario base real, con todos los decimales y sin descuento 
		$product_price_tax_excl = floatval($row['product_price']);
		$qty = floatval($row['product_quantity']);
		$tax = floatval($row['rate']);
		
		// Incorporado a partir de 2.5.2, 2.7.1, 2.6.2
		$unit_base_orig = Tools::ps_round($product_price_tax_excl, 2);
		$unit_base = floatval($row['unit_price_tax_excl']);
		$unit_base_disc = $unit_base_orig - $unit_base;
		$unit_tax = floatval($row['unit_amount']); // floor($after_discount_price_tax_incl * (1 - 100/(100+$tax)) * 100) / 100 // No se redondea, se trunca para pagar menos impuestos = row['unit_price_tax_incl'] - row['unit_price_tax_excl'] = row['unit-amount']
		
		$unit_price_orig = Tools::ps_round($product_price_tax_excl*(1 + $tax/100), 2);
		$unit_price = floatval($row['unit_price_tax_incl']); // $unit_base + $unit_tax;
		$unit_price_disc = $unit_price_orig - $unit_price;
		$discount_perc = floatval($row['reduction_percent']);
		
		$line_price = floatval($row['total_price_tax_incl']); // after_discount_total_tax_incl
		$line_tax = floatval($row['total_amount']); // $unit_tax * $qty; // = $row['total_amount']
		
		return array(
			'order' => intval($row['id_order']),
			// invoice => intval($row['id_order_invoice']),
			'ref' => $is_array? $dto_id[0] : $dto_id,
		    'name' => $this->getTextMultilang($row), // $row['product_name'],
			'size' => $is_array? $dto_id[1] : '',
			'color' => $is_array? $dto_id[2] : '',

			'quantity' => $qty,
			'tax' => $tax,
				
			'unit_base_orig' => $unit_base_orig,
			'unit_base_disc' => $unit_base_disc,
			'discount_perc' => $discount_perc,
			// unit_base => unit_base_orig - unit_base_disc,
			// line_base = unit_base * quantity
			'unit_tax_imp' => $unit_tax, // unit_tax
				
			// Se usaban antes de 2.5.2, 2.7.1, 2.6.2 ¿Harán falta después? Los conservamos por compat con 2.X
			'price' => $unit_price_orig,
			'discount' => $unit_price_disc,
			'total' => $line_price, // line_price = unit_price * quantity
			'tax_imp' => $line_tax, // line_tax = unit_tax * quantity
		);
	}
}

class Fs2psAttributeExtractor extends Fs2psExtractor
{
    protected $selection;
    protected $rMatcher;
    protected $rNextId;
    
    protected function reloadCfg() {
        parent::reloadCfg();
        
        $selection_cfg_name = strtoupper($this->name).'_REMOTE_SELECTION';
        $selection = $this->task->cfg->get($selection_cfg_name);
        if (!empty($selection) && is_array($selection)) {
            $selection = array_filter($selection, 'is_int');
        }
        if (empty($selection)) {
            throw new Fs2psServerFatalException('No se indicó '.$selection_cfg_name);
        }
        $this->selection = $selection;
    }
    
    public static function createRMatcher($task, $name) {
        return new Fs2psDto2RowMatcher('r'.$name, 'product_attribute', 'id_product_attribute', array('ref'), true);
    } 
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name);
        
        $this->rMatcher = self::createRMatcher($task, $name);
        $this->rNextId = Fs2psTools::dbValue('
            select max(cast(dto_id as unsigned)) 
            from @DB_fs2ps_match 
            where `table`=\'product_attribute\' and entity=\'r'.$name.'\'
        ') + 1;  
    }
    
    public function getSelection() { return $this->selection; }
    public function getRMatcher() { return $this->rMatcher; }
    
    protected function buildSql()
    {
        $id_default_lang = Configuration::get('PS_LANG_DEFAULT');
        return '
            select 
            distinct 
                -- GROUP_CONCAT(a.id_attribute_group order by a.id_attribute_group), 
                GROUP_CONCAT(a.id_attribute order by a.id_attribute_group) as id,  
                GROUP_CONCAT(al.name order by a.id_attribute_group SEPARATOR \'x\' ) as name,
                GROUP_CONCAT(a.color order by a.id_attribute_group) as color
            from @DB_product p
            inner join @DB_product_attribute pa on pa.id_product=p.id_product
            inner join @DB_product_attribute_combination pac on pac.id_product_attribute=pa.id_product_attribute
            inner join @DB_attribute a on a.id_attribute=pac.id_attribute
            inner join @DB_attribute_lang al on al.id_attribute=a.id_attribute and id_lang='.$id_default_lang.'
            where p.active=1 and a.id_attribute_group in ('.join($this->selection, ',').')
            group by pa.id_product_attribute
		';
    }
    
    protected function row2dto($row)
    {
        $rMatcher = $this->rMatcher;
        $ref = $rMatcher->_rowIdFromDtoId($row['id']);
        if (empty($ref)) {
            $ref = $this->rNextId;
            $this->rNextId++;
        }
        $rMatcher->updateMatch($row['id'], $ref);
        
        $dto = array( 'ref' => $ref, 'name' => $row['name'] );
        
        if (strlen($row['color'])>0 && $row['color'][0]=='#') {
            $dto['colour'] = substr(explode(',', $row['color'])[0], 1);
        }
        
        return $dto;
        
    }
}

class Fs2psCategoryExtractor extends Fs2psExtractor
{
    
    protected $rMatcher;
    protected $rNextId;
    
    public static function createRMatcher($task, $name) {
        return new Fs2psDto2RowMatcher('r'.$name, 'category', 'id_category', array('ref'), true);
    }
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name);
        
        $this->rMatcher = self::createRMatcher($task, $name);
        $this->rNextId = Fs2psTools::dbValue('
            select max(cast(dto_id as unsigned))
            from @DB_fs2ps_match
            where `table`=\'category\' and entity=\'r'.$name.'\'
        ') + 1;
    }
    
    public function getRMatcher() { return $this->rMatcher; }
    
    protected function row2dto($row)
    {
        $rMatcher = $this->rMatcher;
        $ref = $rMatcher->_rowIdFromDtoId($row['id']);
        if (empty($ref)) {
            $ref = $this->rNextId;
            $this->rNextId++;
        }
        $rMatcher->updateMatch($row['id'], $ref);
        
        return array(
            'ref' => $ref,
            'name' => $row['name'],
            'parent' => null,
        );
    }
    
}

class Fs2psSectionExtractor extends Fs2psCategoryExtractor
{
    protected function buildSql()
    {
        $id_default_lang = Configuration::get('PS_LANG_DEFAULT');
        return '
            select distinct pc.id_category as id, pcl.name, null as id_parent
            from @DB_product p 
            inner join @DB_category_product cp on cp.id_product=p.id_product
            inner join @DB_category c on c.id_category=cp.id_category
            left join @DB_category cc on cc.id_parent=c.id_category
            left join @DB_category pc on pc.id_category=c.id_parent and pc.is_root_category=0
            left join @DB_category_lang pcl on pcl.id_category=pc.id_category and pcl.id_lang='.$id_default_lang.'
            where cc.id_category is null and pc.id_category is not null and p.active=1
            ORDER BY id_parent, id
        ';
    }
}

class Fs2psFamilyExtractor extends Fs2psCategoryExtractor
{
    protected $sectionRMatcher;
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name);
        $this->sectionRMatcher = $task->getExtractor('sections')->getRMatcher();
    }
    
    protected function buildSql()
    {
        $id_default_lang = Configuration::get('PS_LANG_DEFAULT');
        return '
            select distinct c.id_category as id, cl.name, pc.id_category as id_parent
            from @DB_product p
            inner join @DB_category_product cp on cp.id_product=p.id_product
            inner join @DB_category c on c.id_category=cp.id_category
            inner join @DB_category_lang cl on cl.id_category=c.id_category and cl.id_lang='.$id_default_lang.'
            left join @DB_category cc on cc.id_parent=c.id_category
            left join @DB_category pc on pc.id_category=c.id_parent and pc.is_root_category=0
            where cc.id_category is null and pc.id_category is not null and p.active=1
            ORDER BY id_parent, id
        ';
    }
    
    protected function row2dto($row)
    {
        $dto = parent::row2dto($row);
        $dto['parent'] = empty($row['id_parent'])? '' : $this->sectionRMatcher->_rowIdFromDtoId($row['id_parent']);
        return $dto;
    }
}

class Fs2psProductExtractor extends Fs2psExtractor
{
    protected $familyRMatcher;
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name);
        $this->familyRMatcher = $task->getExtractor('families')->getRMatcher();
    }
    
    protected function buildSql()
    {
        $id_default_lang = Configuration::get('PS_LANG_DEFAULT');
        return '
            select 
                p.id_product,
                p.date_add,
                p.date_upd,
                pl.name,
                p.active,
                pl.description as longdescrip,
                pl.description_short as descrip,
                p.price,
                -- p.wholesale_price,
                p.reference as sku, -- pm.dto_id as sku, 
                sa.quantity as stock, -- s.usable_quantity as stock,
                t.rate as tax_class,
                id_family,
                id_section
                -- weight, width, length, height
            from @DB_product p
            -- inner join @DB_fs2ps_match pm on pm.row_id=p.id_product and pm.`table`=\'product\'
            inner join @DB_product_shop ps on ps.id_product=p.id_product
            inner join @DB_product_lang pl on pl.id_product=p.id_product and pl.id_lang='.$id_default_lang.'
            inner join @DB_stock_available sa on sa.id_product=p.id_product and sa.id_product_attribute=0
            inner join @DB_tax_rules_group trg on trg.id_tax_rules_group=p.id_tax_rules_group
            inner join @DB_country c on c.iso_code=\'ES\'
            inner join @DB_tax_rule tr on tr.id_tax_rules_group=p.id_tax_rules_group and tr.id_country=c.id_country
            inner join @DB_tax t on t.id_tax=tr.id_tax
            left join (
                select cp.id_product, min(c.id_category) as id_family, min(pc.id_category) as id_section
                from @DB_category_product cp
                inner join @DB_category c on c.id_category=cp.id_category
                left join @DB_category cc on cc.id_parent=c.id_category
                left join @DB_category pc on pc.id_category=c.id_parent and pc.is_root_category=0
                where cc.id_category is null
                group by cp.id_product
            ) ca on ca.id_product=p.id_product
            where p.active=1
        ';
    }

    protected function row2dto($row)
    {
        if (!isset($row['sku']) || (empty($row['sku']) && $row['sku']!=='0')) { // empty de string '0' se evalúa a true!!!??
            $row['sku'] = $row['id_product'];
            Fs2psTools::dbUpdate('product', array('reference'=>$row['sku']), array('id_product'=>$row['id_product']));
        }
        
        $family = empty($row['id_family'])? '' : $this->familyRMatcher->_rowIdFromDtoId($row['id_family']);
        
        $dto = array(
            'ref' => $row['sku'],
            'name' => $row['name'],
            'created' => $row['date_add'],
            'updated' => $row['date_upd'],
            'enabled' => $row['active']? true : false,
            'price_base' => floatval($row['price']),
            //'cost' => floatval($row['wholesale_price']),
            'stock' => floatval($row['stock']),
            'tax_class' => strval(floatval($row['tax_class'])),
            'descrip' => $row['descrip'],
            'longdescrip' => $row['longdescrip'],
            'family' => empty($row['id_family'])? '' : $family
        );
        
        return $dto;
    }
    
}

class Fs2psSizeColourCombinationExtractor extends Fs2psExtractor
{
    protected $sizeExtractor;
    protected $colourExtractor;
    protected $sizeRMatcher;
    protected $colourRMatcher;
    
    public function __construct($task, $name)
    {
        $this->sizeExtractor = $task->getExtractor('sizes');
        $this->colourExtractor = $task->getExtractor('colours');
        $this->sizeRMatcher = $this->sizeExtractor->getRMatcher();
        $this->colourRMatcher = $this->colourExtractor->getRMatcher();
        
        parent::__construct($task, $name);
    }
    
    protected function buildSql()
    {
        return '
            select 
                min(pa.id_product_attribute) as id_product_attribute,
                min(pa.id_product) as id_product,
                min(pa.reference) as sku,
                min(p.reference) as parent_sku,
                GROUP_CONCAT(at.id_attribute order by at.id_attribute_group) as size,
                GROUP_CONCAT(ac.id_attribute order by ac.id_attribute_group) as colour,
                min(pa.price) as price,
                min(sa.quantity) as stock,
                min(t.rate) as tax_class
            from @DB_product p
            inner join @DB_product_attribute pa on pa.id_product=p.id_product
            inner join @DB_product_attribute_combination pac on pac.id_product_attribute=pa.id_product_attribute
            inner join @DB_stock_available sa on sa.id_product=p.id_product and sa.id_product_attribute=pa.id_product_attribute
            inner join @DB_tax_rules_group trg on trg.id_tax_rules_group=p.id_tax_rules_group
            inner join @DB_country c on c.iso_code=\'ES\'
            inner join @DB_tax_rule tr on tr.id_tax_rules_group=p.id_tax_rules_group and tr.id_country=c.id_country
            inner join @DB_tax t on t.id_tax=tr.id_tax
            left join @DB_attribute at on at.id_attribute=pac.id_attribute and at.id_attribute_group in ('.join($this->sizeExtractor->getSelection(), ',').')
            left join @DB_attribute ac on ac.id_attribute=pac.id_attribute and ac.id_attribute_group in ('.join($this->colourExtractor->getSelection(), ',').')
            where p.active=1
            group by pa.id_product_attribute
		';
    }
    
    protected function row2dto($row)
    {
        $size = empty($row['size'])? '' : $this->sizeRMatcher->_rowIdFromDtoId($row['size']);
        $colour = empty($row['colour'])? '' : $this->colourRMatcher->_rowIdFromDtoId($row['colour']);
        
        if (!isset($row['sku']) || (empty($row['sku']) && $row['sku']!=='0')) { // empty de string '0' se evalúa a true!!!??
            $row['sku'] = $row['parent_sku'].'_'.$size.'_'.$colour;
            Fs2psTools::dbUpdate('product_attribute', array('reference'=>$row['sku']), array('id_product_attribute'=>$row['id_product_attribute']));
        }
        
        $dto = array(
            'ref' => $row['sku'],
            'product' => $row['parent_sku'],
            'size' => $size,
            'colour' => $colour,
            'price_base' => floatval($row['price']),
            //'cost' => floatval($row['wholesale_price']),
            'stock' => floatval($row['stock']),
            'tax_class' => strval(floatval($row['tax_class'])),
        );
        
        return $dto;
    }

}
