<?php
/**
* LICENCIA
*
* Este programa se propociona "tal cual", sin garantía de ningún tipo más allá del soporte
* pactado a la hora de adquirir el programa.
*
* En ningún caso los autores o titulares del copyright serán responsables de ninguna
* reclamación, daños u otras responsabilidades, ya sea en un litigio, agravio o de otro
* modo, que surja de o en conexión con el programa o el uso u otro tipo de acciones
* realizadas con el programa.
*
* Este programa no puede modificarse ni distribuirse sin el consentimiento expreso del autor.
*
*    @author    Carlos Fillol Sendra <festeweb@festeweb.com>
*    @copyright 2014 Fes-te web! - www.festeweb.com
*    @license   http://www.festeweb.com/static/licenses/fs2ps_1.1.0.txt
*/

include_once(dirname(__FILE__).'/Fs2psTools.php');
include_once(dirname(__FILE__).'/Fs2psUpdater.php');
include_once(dirname(__FILE__).'/Fs2psDto2RowRefMatcher.php');


class Fs2psProduct extends Product
{
}

class Fs2psCategory extends Category
{
	public function add($autodate = true, $null_values = false)
	{
		$this->_adding = true;
		parent::add($autodate, $null_values);
	}
	
	public function update($null_values = false)
	{
		$this->_adding = false;
		parent::update($null_values);
	}
	
	public function updateGroup($list)
	{
		if ($this->_adding)
			parent::updateGroup($list);
	}
}

class Fs2psManufacturer extends Manufacturer
{
}

class Fs2psSupplier extends Supplier
{
}

class Fs2psGroup extends Group
{
}

class Fs2psAttributeGroup extends AttributeGroup
{
}

class Fs2psAttribute extends Attribute
{
}

class Fs2psCombination extends Combination
{
}

class Fs2psFeature extends Feature
{
}

class Fs2psFeatureValue extends FeatureValue
{
}

class Fs2psSpecificPrice extends SpecificPrice
{

    public static function fs2psGetSpecificPrice($id_product, $id_product_attribute, $id_shop, $id_group, $id_customer, $from_quantity=1)
	{
		return Fs2psTools::dbRow('
				SELECT id_specific_price
				FROM `'._DB_PREFIX_.'specific_price`
				WHERE `id_product`='.(int)$id_product.'
				AND `id_product_attribute`='.(int)$id_product_attribute.'
				AND `id_shop`='.(int)$id_shop.'
				AND `id_currency`=0
				AND `id_country`=0
				AND `id_group`='.(int)$id_group.'
                AND `id_customer`='.(int)$id_customer.'
				AND `from` = \'0000-00-00 00:00:00\'
				AND	`to` = \'0000-00-00 00:00:00\'
				AND `from_quantity`='.(int)$from_quantity.'
				ORDER BY `id_product_attribute` DESC'
		);
	}
	
	public static function addOrUpdateProductPriceRate($price, $id_product, $id_product_attribute=0, $id_shop=0, $id_group=0, $id_customer=0, $reduction=0.0, $reduction_type='amount', $from_quantity=1)
	{
	    $oldSpecPrice = self::fs2psGetSpecificPrice($id_product, $id_product_attribute, $id_shop, $id_group, $id_customer, $from_quantity);
		$id_specific_price = $oldSpecPrice? (int)$oldSpecPrice['id_specific_price'] : null;
		$specPrice = new SpecificPrice($id_specific_price);
		$specPrice->id_product = $id_product;
		$specPrice->id_product_attribute = $id_product_attribute;
		$specPrice->id_shop = $id_shop;
		$specPrice->id_currency = 0;
		$specPrice->id_country = 0;
		$specPrice->id_group = $id_group;
		$specPrice->id_customer = $id_customer;
		$specPrice->price = $price;
		$specPrice->from_quantity = $from_quantity;
		$specPrice->reduction = $reduction;
		$specPrice->reduction_type = $reduction_type;
		// reduction_tax no existe en ver. viejas de la 1.6 (ej: 1.6.0.9). Existe al menos desde la 1.6.1.18.
		if (property_exists($specPrice, 'reduction_tax')) $specPrice->reduction_tax = 0;  
		
		$specPrice->from = '0000-00-00 00:00:00';
		$specPrice->to = '0000-00-00 00:00:00';
		$specPrice->save();
	}
	
	public static function resetProductManagedPriceRates($id_product, $id_product_attribute=0, $id_shop=null, $id_group=null, $id_customer=0, $from_quantity=1)
	{
		$shop_where = $id_shop===null? '' : ' sp.id_shop='.(int)$id_shop.' AND ';
		$group_join = $id_group===null ? ' inner join `@DB_fs2ps_match` m on `table`=\'group\' and m.row_id=sp.id_group ' : '';
		$group_where = $id_group===null? '' : ' sp.id_group='.(int)$id_group.' AND ';
		$attribute_where = $id_product_attribute===null? '' : ' sp.id_product_attribute='.(int)$id_product_attribute.' AND ';
		$customer_where = $id_customer===null? '' : ' sp.id_customer='.(int)$id_customer.' AND ';
		$from_quantity_where = $from_quantity===null? '' : ' sp.from_quantity='.(int)$from_quantity.' AND ';
		Fs2psTools::dbExec('
		    DELETE sp
			FROM `@DB_specific_price` as sp
                '.$group_join.'
            WHERE
				'.$shop_where.'
				'.$group_where.'
            	sp.id_product='.(int)$id_product.' AND
                '.$customer_where.'
                '.$attribute_where.'
                '.$from_quantity_where.'
                sp.id_currency=0 AND sp.id_country=0 
		');
	}
	
	public static function resetProductManagedGroupPriceRates($conserve_shop_group_qty, $affected_shops, $allow_group0_deletion, $id_product=null, $id_product_attribute=null)
	{
		$where = array(
			'sp.id_currency=0 AND sp.id_country=0 AND sp.id_customer=0',
			'sp.id_shop in ('.implode(', ', $affected_shops).')',
		);
		if ($conserve_shop_group_qty) {
		    if (sizeof($conserve_shop_group_qty[0])==3) { 
		        // managing from_quantity
		        $where[] = '(sp.id_shop, sp.id_group, sp.from_quantity) not in ('.implode(', ', array_map(function ($e) { return '('.$e[0].', '.$e[1].', '.$e[2].')'; }, $conserve_shop_group_qty)).')';
		    } else {
		        // not managing from_quantity
		        $where[] = '(sp.id_shop, sp.id_group) not in ('.implode(', ', array_map(function ($e) { return '('.$e[0].', '.$e[1].')'; }, $conserve_shop_group_qty)).')';
                $where[] = 'sp.from_quantity=1';
		    }
		}
		if ($id_product!=null) $where[] = 'sp.id_product='.(int)$id_product;
		if ($id_product_attribute!=null) $where[] = 'sp.id_product_attribute='.(int)$id_product_attribute;
		if ($allow_group0_deletion!=null) $where[] = '((sp.id_group=0 and m.row_id is null) or (sp.id_group<>0 and m.row_id is not null))';
		
		Fs2psTools::dbExec('
		    DELETE sp
			FROM `@DB_specific_price` as sp
				'.($allow_group0_deletion? 'left' : 'inner').' join `@DB_fs2ps_match` m on `table`=\'group\' and m.row_id=sp.id_group
				'.($id_product==null? 'inner join `@DB_fs2ps_match` mp on mp.`table`=\'product\' and mp.row_id=sp.id_product and mp.uploaded=1' : '').'
            WHERE '.implode(' AND ', $where)
		);
	}
	
	public static function resetProductManagedCustomerPriceRates($conserve_customers_qty, $affected_shops, $id_product=null, $id_product_attribute=null)
	{
	    $where = array(
	        'sp.id_currency=0 AND sp.id_country=0',
	        'sp.id_shop in ('.implode(', ', $affected_shops).')'
	    );
	    
	    if ($conserve_customers_qty) {
	        if (is_array($conserve_customers_qty[0])) {
	            // managing from_quantity
	            $where[] = '(sp.id_customer, sp.from_quantity) not in ('.implode(', ', array_map(function ($e) { return '('.$e[0].', '.$e[1].')'; }, $conserve_customers_qty)).')';
	        } else {
	            // not managing from_quantity
	            $where[] = 'sp.id_customer not in ('.implode(', ', $conserve_customers_qty).')';
	            $where[] = 'sp.from_quantity=1';
	        }
	    }
	    if ($id_product!=null) $where[] = 'sp.id_product='.(int)$id_product;
	    if ($id_product_attribute!=null) $where[] = 'sp.id_product_attribute='.(int)$id_product_attribute;
	    
	    Fs2psTools::dbExec('
		    DELETE sp
			FROM `@DB_specific_price` as sp
				inner join `@DB_fs2ps_match` m on `table`=\'customer\' and m.row_id=sp.id_customer
				'.($id_product==null? 'inner join `@DB_fs2ps_match` mp on mp.`table`=\'product\' and mp.row_id=sp.id_product and mp.uploaded=1' : '').'
            WHERE '.implode(' AND ', $where)
	    );
	}
	
	public static function resetProductManagedFromQuantityPriceRates($from_quantities, $affected_shops, $id_product=null, $id_product_attribute=null)
	{
	    $where = array(
	        'sp.id_currency=0 AND sp.id_country=0',
	        'sp.id_shop in ('.implode(', ', $affected_shops).')'
	    );
	    if ($from_quantities) {
	        $where[] = 'sp.from_quantity not in ('.implode(', ', $from_quantities).')';
	    }
	    if ($id_product!=null) $where[] = 'sp.id_product='.(int)$id_product;
	    if ($id_product_attribute!=null) $where[] = 'sp.id_product_attribute='.(int)$id_product_attribute;
	    
	    Fs2psTools::dbExec('
		    DELETE sp
			FROM `@DB_specific_price` as sp
				'.($id_product==null? 'inner join `@DB_fs2ps_match` mp on mp.`table`=\'product\' and mp.row_id=sp.id_product and mp.uploaded=1' : '').'
            WHERE '.implode(' AND ', $where)
	        );
	}
	
}

class Fs2psCustomer extends Customer
{
    
    public static function resetFs2psGroups($idCustomer)
    {
        return Fs2psTools::dbExec('
            DELETE cg
            FROM `@DB_customer_group` cg
            INNER JOIN `@DB_fs2ps_match` m on m.entity=\'price_rates\' and m.row_id=cg.id_group
            WHERE cg.id_customer='.(int)$idCustomer.'

        ');
    }
    
    public static function getFs2psGroups($idCustomer)
    {
        return Fs2psTools::dbSelect('
			SELECT cg.id_group
            FROM `@DB_customer_group` cg
            INNER JOIN `@DB_fs2ps_match` m on m.entity=\'price_rates\' and m.row_id=cg.id_group
            WHERE cg.id_customer='.(int)$idCustomer.'
        ');
    }
    
}

class Fs2psAddress extends Address
{
    
}

class Fs2psOrder extends Order
{
    public function getFs2psInvoicesPdfsSubfolder()
    {
        $reference = $this->reference;
        $sf1 = substr($reference, 0, 2);
        $sf2 = substr($reference, 2, 2);
        return ($sf1? $sf1 : 'AA').'/'.($sf2? $sf2 : 'AA');
    }
}

