<?php
/**
 * LICENCIA
 *
 * Este programa se propociona "tal cual", sin garantía de ningún tipo más allá del soporte
 * pactado a la hora de adquirir el programa.
 *
 * En ningún caso los autores o titulares del copyright serán responsables de ninguna
 * reclamación, daños u otras responsabilidades, ya sea en un litigio, agravio o de otro
 * modo, que surja de o en conexión con el programa o el uso u otro tipo de acciones
 * realizadas con el programa.
 *
 * Este programa no puede modificarse ni distribuirse sin el consentimiento expreso del autor.
 *
 *    @author    Carlos Fillol Sendra <festeweb@festeweb.com>
 *    @copyright 2014 Fes-te web! - www.festeweb.com
 *    @license   http://www.festeweb.com/static/licenses/fs2ps_1.1.0.txt
 */

include_once(dirname(__FILE__).'/Fs2psTools.php');
include_once(dirname(__FILE__).'/Fs2psUpdater.php');
include_once(dirname(__FILE__).'/Fs2psDto2RowRefMatcher.php');
include_once(dirname(__FILE__).'/Fs2psObjectModels.php');

class Fs2psActivableContentUpdater extends Fs2psUpdater
{
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array();
        
        if ($this->enable)
            $row['active'] = 1;
            elseif (!$exists)
            $row['active'] = 0;
            
            if (!$exists || !$this->noover_content)
            {
                $all_langs = !$exists;
                
                if (isset($dto['name']))
                {
                    $name = $dto['name'];
                    $row['name'] = Fs2psTools::multiLangField($name, $all_langs);
                    $row['link_rewrite'] = Fs2psTools::multiLangLinkRewrite($name, $all_langs);
                }
            }
            return $row;
    }
}

class Fs2psCategoryUpdater extends Fs2psActivableContentUpdater
{
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'category');
    }
    
    protected function reloadCfg() {
        parent::reloadCfg();
        $cfg = $this->task->cfg;
        $this->noover_parent = $cfg->get('NOOVER_CATEGORIES_PARENT', false);
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = parent::dto2row($dto, $idx, $exists, $oldRowId);
        
        if (!$exists || !$this->noover_parent)
        {
            $id_parent = !empty($dto['parent'])? $this->matcher->rowIdFromDtoId($dto['parent']) : null;
            $row['id_parent'] =$id_parent? $id_parent : Configuration::get('PS_HOME_CATEGORY');
        }
        
        return $row;
    }
}

class Fs2psNoMultilangActivableContentUpdater extends Fs2psUpdater
{
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array();
        
        if ($this->enable)
            $row['active'] = 1;
            elseif (!$exists)
            $row['active'] = 0;
            
            if (!$exists || !$this->noover_content)
            {
                if (isset($dto['name']))
                {
                    $name = $dto['name'];
                    $row['name'] =  $name;
                    $row['link_rewrite'] = $name;
                }
            }
            return $row;
    }
}

class Fs2psManufacturerUpdater extends Fs2psNoMultilangActivableContentUpdater
{
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'manufacturer');
    }
}

class Fs2psSupplierUpdater extends Fs2psNoMultilangActivableContentUpdater
{
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'supplier');
    }
}

class Fs2psGroupUpdater extends Fs2psUpdater
{
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'group');
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array();
        
        if (!$exists)
        {
            # Inicializamos sólo al crear
            $row['price_display_method'] = 1;
        }
        
        if (!$exists || !$this->noover_content)
        {
            $all_langs = !$exists;
            
            if (isset($dto['name']))
            {
                $name = $dto['name'];
                $row['name'] = Fs2psTools::multiLangField($name, $all_langs);
            }
        }
        return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        return parent::insertOrUpdate($row, $exists, $oldRowId);
    }
}

class Fs2psAttributeGroupUpdater extends Fs2psUpdater
{
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'attribute_group');
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array(
            'is_color_group' => $dto['is_color_group'],
            'group_type' => $dto['is_color_group'] ? 'color' : 'select'
        );
        
        if (!$exists || !$this->noover_content)
        {
            $all_langs = !$exists;
            
            if (isset($dto['name']))
                $row['name'] = Fs2psTools::multiLangField($dto['name'], $all_langs);
                
                if (isset($dto['public_name']))
                    $row['public_name'] = Fs2psTools::multiLangField($dto['public_name'], $all_langs);
                    
                    $row['position'] = $idx;
        }
        return $row;
    }
}

class Fs2psAttributeUpdater extends Fs2psUpdater
{
    
    protected $group_row_id;
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'attribute');
    }
    
    public function getGroupRowId()
    {
        if (empty($this->group_row_id))
        {
            $dto_id = strtoupper($this->name);
            $attribute_group_upd = $this->task->getUpdater('attribute_groups');
            $this->group_row_id = $attribute_group_upd->matcher->rowIdFromDtoId($dto_id);
        }
        return $this->group_row_id;
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array(
            'id_attribute_group' => $this->getGroupRowId(),
        );
        
        if (!$exists || !$this->noover_content)
        {
            $all_langs = !$exists;
            
            if (isset($dto['name']))
                $row['name'] = Fs2psTools::multiLangField($dto['name'], $all_langs);
                
                if (isset($dto['color'])) $row['color'] = !empty($dto['color']) ? $dto['color'] : null;
                if (!$exists) $row['position'] = $idx;
        }
        
        return $row;
    }
}


class Fs2psFeatureUpdater extends Fs2psUpdater
{
    
    public function __construct($task, $name)
    {
        parent::__construct($task, $name, 'feature');
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array();
        
        if (!$exists || !$this->noover_content)
        {
            $all_langs = !$exists;
            
            if (isset($dto['name']))
                $row['name'] = Fs2psTools::multiLangField($dto['name'], $all_langs);
                
                if (!$exists) $row['position'] = $idx;
        }
        return $row;
    }
    
}


class Fs2psSKUUpdater extends Fs2psUpdater
{
    protected $price_rates_upd;
    protected $customer_upd;
    protected $nogroup_specific_prices;
    protected $shop_group_pairs = null;
    protected $affected_shops = null;
    
    public function __construct($task, $name, $table, $dto_id_fields = array('ref'), $matcher = null)
    {
        parent::__construct($task, $name, $table, $dto_id_fields, $matcher);
        $this->price_rates_upd = $task->getUpdater('price_rates');
        $this->customer_upd = new Fs2psCustomerUpdater($task, 'customers');
    }
    
    protected function reloadCfg() {
        parent::reloadCfg();
        $cfg = $this->task->cfg;
        $this->nogroup_specific_prices = $cfg->get('NOGROUP_SPECIFIC_PRICES', False);
        $this->managing_price_by_customer = $cfg->get('MANAGING_PRICE_BY_CUSTOMER');
        $this->managing_price_from_quantity = $cfg->get('MANAGING_PRICE_FROM_QUANTITY');
        $this->same_group_price_rates = $cfg->get('SAME_GROUP_PRICE_RATES');
    }
    
    protected function dto_price_id_group($price)
    {
        return isset($price['group'])? $this->price_rates_upd->matcher->rowIdFromDtoId($price['group']) : 0;
    }
    
    protected function dto_price_id_customer($price)
    {
        return isset($price['customer'])? $this->customer_upd->matcher->rowIdFromDtoId($price['customer']) : 0;
    }
    
    protected function pricesByShop($prices)
    {
        $prices_by_shop = array();
        foreach($prices as $price)
        {
            $id_shop = empty($price['shop'])? 0 : $price['shop'];
            if (!array_key_exists($id_shop, $prices_by_shop)) {
                $prices_by_shop[$id_shop] = array();
            }
            array_push($prices_by_shop[$id_shop], $price);
        }
        return $prices_by_shop;
    }
    
    protected function removeUnusedFromQuantityPriceRates($prices_by_shop, $id_product, $id_product_attribute=0) {
        $from_quantities = array();
        foreach($prices_by_shop as $prices) {
            foreach($prices as $price) {
                if (!empty($price['fromqty']) && empty($price['customer']) && empty($price['group'])) {
                    $from_quantities[] = $price['fromqty'];
                }
            }
        }
        Fs2psSpecificPrice::resetProductManagedFromQuantityPriceRates($from_quantities, $this->affected_shops, $id_product, $id_product_attribute);
    }
    
    protected function removeUnusedGroupPriceRates($prices_by_shop, $id_product, $id_product_attribute=0) {
        $conserve_shop_group_pairs = array();
        $shop_idx = 0;
        foreach($prices_by_shop as $id_shop=>$prices) {
            foreach($prices as $idx=>$price) {
                $es_precio_por_defecto_global = $shop_idx==0 && $idx==0;
                $reduction_field = isset($price['disp'])? 'disp' : 'dis';
                $reduction = empty($price[$reduction_field])? 0.0 : $price[$reduction_field]; // Siempre sin IVA
                $fromqty = !$this->managing_price_from_quantity || empty($price['fromqty'])? 1 : (int)$price['fromqty'];
                if (!empty($price['group']) && (!$es_precio_por_defecto_global || $reduction || $fromqty>1)) {
                    // Excluimos de $conserve_shop_group_pairs el precio por defecto global para eliminarlo como precio específico
                    $id_group = $this->dto_price_id_group($price);
                    if ($id_group) {
                        if ($this->managing_price_from_quantity) {
                            $conserve_shop_group_pairs[] = array($id_shop, $id_group, $fromqty);
                        } else {
                            $conserve_shop_group_pairs[] = array($id_shop, $id_group);
                        }
                    }
                }
            }
            $shop_idx++;
        }
        Fs2psSpecificPrice::resetProductManagedGroupPriceRates($conserve_shop_group_pairs, $this->affected_shops, true, $id_product, $id_product_attribute);
    }
    
    protected function removeUnusedCustomerPriceRates($prices_by_shop, $id_product, $id_product_attribute=0) {
        $conserve_customers = array();
        foreach($prices_by_shop as $prices) {
            foreach($prices as $price) {
                if (!empty($price['customer'])) {
                    $id_customer = $this->dto_price_id_customer($price);
                    if ($id_customer) {
                        if ($this->managing_price_from_quantity) {
                            $conserve_customers[] = array($id_customer, empty($price['fromqty'])? 1 : (int)$price['fromqty']);
                        } else {
                            $conserve_customers[] = $id_customer;
                        }
                    }
                }
            }
        }
        Fs2psSpecificPrice::resetProductManagedCustomerPriceRates($conserve_customers, $this->affected_shops, $id_product, $id_product_attribute);
    }
    
    protected function insertOrUpdatePrice($row, $id_product, $id_product_attribute=0)
    {
        if ($this->nogroup_specific_prices) { // Nuevas versiones con soporte de tarifas por grupo y multitienda
            if (empty($row['_prices']))
            {
                // No se indicaron precios específicos. El precio será el indicado en prod/combi. Eliminamos reglas.
                // No es necesario eliminarlo aquí. Se eliminará en el onProcessEnd
            }
            else
            {
                $prices_by_shop = $this->pricesByShop($row['_prices']);
                if ($this->managing_price_from_quantity) {
                    $this->removeUnusedFromQuantityPriceRates($prices_by_shop, $id_product, $id_product_attribute);
                }
                if ($this->managing_price_by_customer) {
                    $this->removeUnusedCustomerPriceRates($prices_by_shop, $id_product, $id_product_attribute);
                }
                if (!$this->same_group_price_rates || $this->managing_price_from_quantity) {
                    $this->removeUnusedGroupPriceRates($prices_by_shop, $id_product, $id_product_attribute);
                }
                
                $shop_idx = 0;
                foreach($prices_by_shop as $id_shop=>$prices)
                {
                    $es_tienda_por_defecto = $shop_idx==0;
                    foreach($prices as $idx=>$price)
                    {
                        if (!$this->managing_price_by_customer && !empty($price['customer'])) continue;
                        
                        if (!empty($price['group']) && !empty($price['customer'])) {
                            $msg = 'WARN: No puede indicar una tarifa que afecte a un grupo ('.$price['group'].') y a un cliente ('.$price['customer'].') a la vez (prod. "'.$row['_dto_id'].'")';
                            $this->task->log($msg);
                            continue;
                        }
                        
                        $id_group = $this->dto_price_id_group($price);
                        if (!empty($price['group']) && empty($id_group)) {
                            $msg = 'WARN: El grupo "'.$price['group'].'" referenciado por "'.$row['_dto_id'].'" no existe';
                            $this->task->log($msg);
                            continue;
                        }
                        
                        $id_customer = $this->dto_price_id_customer($price);
                        if (!empty($price['customer']) && empty($id_customer)) {
                            $msg = 'WARN: El cliente "'.$price['customer'].'" referenciado por "'.$row['_dto_id'].'" no existe';
                            $this->task->log($msg);
                            continue;
                        }
                        
                        $reduction_field = isset($price['disp'])? 'disp' : 'dis';
                        $reduction = empty($price[$reduction_field])? 0.0 : $price[$reduction_field]; // Siempre sin IVA
                        if ($reduction_field=='disp') $reduction = $reduction/100.0;
                        $saleprice = $price['price']; // Sin IVA
                        
                        $fromqty = !$this->managing_price_from_quantity || empty($price['fromqty'])? 1 : (int)$price['fromqty'];
                        
                        $es_precio_por_defecto_tienda = $idx==0 && ($shop_idx==0 || $id_shop>0);
                        $es_precio_por_defecto_global = $shop_idx==0 && $idx==0;
                        
                        // Si es la primera tarifa y no tiene descuento, no se indica, en caso contrario, sí
                        // Si no es primera tarifa, hay que crear o actualizar tenga o no descuento
                        if (!$es_precio_por_defecto_global || $reduction || $fromqty>1) {
                            // La primera tarifa asociada a la tienda será el PVP para esa tienda (id_group==0).
                            Fs2psSpecificPrice::addOrUpdateProductPriceRate(
                                $reduction && $es_precio_por_defecto_global? -1 : $saleprice, 	// XXX: // Evitamos bug http://forge.prestashop.com/browse/PSCSX-8512
                                $id_product,
                                $id_product_attribute, 											// Con $id_product_attribute>0 evitamos q aparezca precio 0 en prod
                                $es_tienda_por_defecto? 0 : $id_shop,
                                $es_precio_por_defecto_tienda? 0 : $id_group,
                                $id_customer,
                                $reduction,
                                $reduction_field=='disp'? 'percentage' : 'amount',
                                $fromqty
                                );
                        } else if ($es_precio_por_defecto_global) { // && !$reduction && $fromqty<=1
                            // Si es la primera tarifa y no tiene descuento, hay que borrar el descuento que pudiera haber anteriormente
                            // Sólo lo hacemos en el caso del precio por defecto por temas de rendimiento. El resto lo haremos globalmente en el onProcessStart/onProcessEnd
                            Fs2psSpecificPrice::resetProductManagedPriceRates($id_product, $id_product_attribute, 0, 0);
                        }
                    }
                    $shop_idx++;
                }
            }
        } else { // Comportamiento por defecto, el de versiones antiguas < 2.2.0. Multitarifa sin multitienda ni rebajas ni precio por cliente ni fromqty
            if (!empty($row['_prices'])) {
                foreach($row['_prices'] as $price)
                {
                    $reduction_field = isset($price['disp'])? 'disp' : 'dis';
                    $reduction = empty($price[$reduction_field])? 0.0 : $price[$reduction_field]; // Siempre sin IVA
                    if ($reduction_field=='disp') $reduction = $reduction/100.0;
                    $saleprice = $price['price']; // Sin IVA
                    $id_group = $this->dto_price_id_group($price);
                    Fs2psSpecificPrice::addOrUpdateProductPriceRate(
                        $saleprice, $id_product, $id_product_attribute, 0, $id_group, 0, $reduction, $reduction_field=='disp'? 'percentage' : 'amount');
                }
            } // else // Se eliminaran precios específicos del prod./combi. en el onProcessEnd
        }
    }
    
    protected function onProcessStart($dtos) {
        if (empty($dtos)) return;
        
        $nogroup_specific_prices = $this->nogroup_specific_prices; // Si true trabajaremos con grupo==0
        
        if (empty($dtos[0]['prices'])) {
            $this->shop_group_pairs = array();
            $this->affected_shops = array(0);
            return;
        } else {
            $prices = $dtos[0]['prices'];
            $shop_group_pairs = array();
            $affected_shops = array(0=>1);
            $prices_by_shop = $this->pricesByShop($prices);
            $shop_idx = 0;
            foreach($prices_by_shop as $id_shop=>$prices) {
                $group_idx = 0;
                $affected_shops[$id_shop] = 1;
                foreach($prices as $price) {
                    // Apuntamos las tarifas que se usan para después eliminar los precios específicos de las tarifas que ya no se usan
                    // Los precios específicos para precios globales (tienda, grupo)==(0, 0) se eliminan individualmente para cada producto/combinación
                    array_push($shop_group_pairs, array(
                        $shop_idx==0? 0 : $id_shop,
                        $nogroup_specific_prices && $group_idx==0? 0 : $this->dto_price_id_group($price)
                    ));
                    $group_idx++;
                }
                $shop_idx++;
            }
            $this->affected_shops = array_keys($affected_shops);
            $this->shop_group_pairs = $shop_group_pairs;
        }
    }
    
    protected function onProcessEnd($dtos) {
        if ($this->affected_shops!==null && $this->same_group_price_rates && !$this->managing_price_from_quantity) {
            // Optimización para borrar las que ya no se usan de una tacada.
            // Sólo es posible si todos los artículos comparten los mismos  grupos de tarifas
            Fs2psSpecificPrice::resetProductManagedGroupPriceRates($this->shop_group_pairs, $this->affected_shops, $this->nogroup_specific_prices);
        }
    }
    
}


class Fs2psProductUpdater extends Fs2psSKUUpdater
{
    protected $categories_upd;
    protected $manufacturers_upd;
    protected $suppliers_upd;
    protected $features_upd;
    
    protected $priorities = null;
    
    protected $noover_categories;
    protected $noover_manufacturer;
    protected $noover_suppliers;
    
    protected $id_tax_rules_group_by_rate;
    protected $dims = array('width', 'height', 'depth', 'weight');
    
    public static function createMatcher($task, $name) {
        $dto_id_fields = array('ref');
        $pm = $task->cfg->get('PRODUCTS_MATCHER');
        $matcher_cls = $pm=='basic'? 'Fs2psDto2RowMatcher' : ($pm=='pref'? 'Fs2psDto2RowPRefMatcher' : 'Fs2psDto2RowRefMatcher');
        return new $matcher_cls($name, 'product', 'id_product', $dto_id_fields, true);
    } 
    
    public function __construct($task, $name)
    {
        $matcher = self::createMatcher($task, $name);
        
        parent::__construct($task, $name, 'product', $matcher->dto_id_fields, $matcher);
        
        $this->categories_upd = $task->getUpdater('categories');
        $this->manufacturers_upd = $task->getUpdater('manufacturers');
        $this->suppliers_upd = $task->getUpdater('suppliers');
        $this->features_upd = $task->getUpdater('features');
        
        $this->loadTaxRulesGroupByRate();
    }
    
    protected function reloadCfg() {
        parent::reloadCfg();
        $cfg = $this->task->cfg;
        $this->noover_categories = $cfg->get('NOOVER_PRODUCTS_CATEGORIES', false);
        $this->noover_manufacturer = $cfg->get('NOOVER_PRODUCTS_MANUFACTURER', false);
        $this->noover_suppliers = $cfg->get('NOOVER_PRODUCTS_SUPPLIERS', false);
    }
    
    protected function loadTaxRulesGroupByRate()
    {
        $id_tax_rules = Fs2psTools::dbSelect('
			SELECT distinct tr.id_tax_rules_group, tr.id_tax, t.rate
			FROM @DB_tax_rule tr, @DB_tax_rules_group trg, @DB_tax t
			where
			    tr.id_tax=t.id_tax and
                trg.id_tax_rules_group=tr.id_tax_rules_group and 
                trg.active=1 and tr.behavior=0
				'.(version_compare(_PS_VERSION_, '1.6.1.0') >= 0? ' and trg.deleted<>1 ' : '').'
		');
        $id_tax_rules_group_by_rate = array();
        foreach ($id_tax_rules as $rule)
        {
            $rate = (string)((float)$rule['rate']);
            $id_tax_rules_group_by_rate[$rate] = $rule['id_tax_rules_group'];
        }
        $this->id_tax_rules_group_by_rate = $id_tax_rules_group_by_rate;
    }
    
    public static function deleteFeatures($id_product)
    {
        // Si fv.custom=1, distintos id_feature_value para distintos id_product.
        //   - Es seguro borrar fv y fvl por que no se va a interferir en otros productos.
        //   - fp lo borraremos en la siguiente sentencia
        // Si fv.custom=0, id_feature_value coincide para los distintos productos cuando se trata del mismo valor
        //   - Sólo debemos borrar feature_product (en la siguiente sentencia)
        Fs2psTools::dbExec('
			delete fv, fvl
            from @DB_fs2ps_match m
            inner join @DB_feature_product fp on fp.id_feature=m.row_id
            left join @DB_feature_value fv on fv.id_feature_value=fp.id_feature_value and fv.custom=1
            left join @DB_feature_value_lang fvl on fvl.id_feature_value=fv.id_feature_value
            where m.`table`=\'feature\' and fp.id_product='.(int)$id_product.'
		');
        Fs2psTools::dbExec('
			delete fp
            from @DB_fs2ps_match m
            inner join @DB_feature_product fp on fp.id_feature=m.row_id
            left join @DB_feature_value fv on fv.id_feature_value=fp.id_feature_value
            left join @DB_feature_value_lang fvl on fvl.id_feature_value=fv.id_feature_value
            where m.`table`=\'feature\' and fp.id_product='.(int)$id_product.'
		');
        
        /* XXX cfillol: Extraído de Product.deleteFeatures, pero es realmente necesario?
         SpecificPriceRule::applyAllRules(array((int)$this->id));
         */
    }
    
    public function createFeatureValue($id_feature, $custom, $value) {
        $id_feature_value = Fs2psTools::dbInsert('feature_value', array( 'id_feature' => $id_feature, 'custom' => $custom));
        $languages = Language::getLanguages(false);
        foreach ($languages as $language) {
            Fs2psTools::dbInsert('feature_value_lang', array('id_feature_value' => $id_feature_value, 'id_lang' => $language['id_lang'], 'value' => $value));
        }
        return $id_feature_value;
    }
    
    public function addFeatures($id_product, $features)
    {
        foreach ($features as $feature) {
            $values = is_array($feature['value'])? $feature['value'] : array($feature['value']);
            if ($feature['custom']) {
                foreach ($values as $value) {
                    $id_feature_value = $this->createFeatureValue($feature['id'], 1, $value);
                    Product::addFeatureProductImport($id_product, $feature['id'], $id_feature_value);
                }
            } else {
                foreach ($values as $value) {
                    $id_feature_value = Fs2psTools::dbValue('
            			select id_feature_value from @DB_feature_value_lang
            			where id_lang='.(int)$this->id_default_lang.' and value=\''.Fs2psTools::dbEscape($value).'\'
            		');
                    if (empty($id_feature_value)) {
                        $id_feature_value = $this->createFeatureValue($feature['id'], 0, $value);
                    }
                    Product::addFeatureProductImport($id_product, $feature['id'], $id_feature_value);
                }
            }
        }
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $rate = (string)($dto['iva'] * 100);
        if (empty($this->id_tax_rules_group_by_rate[$rate]))
            throw new Fs2psException(
                'No se definió en la tienda un IVA del '.$rate.'% (prod. "'.$this->matcher->dtoIdToStrFromDto($dto).'")'
                );
            
            $row = array(
                // Do not include id here, it will be automatically included
                'reference' => $this->matcher->referenceFromDto($dto),
                'wholesale_price' => isset($dto['cost']) ? $dto['cost'] : 0,
                'id_tax_rules_group' => ($this->id_tax_rules_group_by_rate[$rate]),
                '_dto_id' => $this->matcher->dtoId($dto),
                '_quantity' => isset($dto['stock']) ? $dto['stock'] : 0,
                '_tax_rate' => $dto['iva']
            );
            
            // Prices. Multiple price rates or simple price.
            if (empty($dto['prices'])) {
                $row['price'] = isset($dto['price']) ? $dto['price'] : 0;
            } else {
                $row['_prices'] = $dto['prices'];
                $row['price'] = $dto['prices'][0]['price'];
            }
            
            if ($this->enable)
                $row['active'] = 1;
                elseif (!$exists)
                $row['active'] = 0;
                
                if (!$exists || !$this->noover_content)
                {
                    $all_langs = !$exists;
                    
                    if (isset($dto['name']))
                    {
                        $name = $dto['name'];
                        $row['name'] = Fs2psTools::multiLangField($name, $all_langs, array('replace'=>array("/\r\n|\r|\n/", '')));
                        $row['link_rewrite'] = Fs2psTools::multiLangLinkRewrite($name, $all_langs);
                        //'meta_title' => '', // name by default
                    }
                    
                    if (isset($dto['descrip']))
                    {
                        $descrip = $dto['descrip'];
                        $row['description_short'] = Fs2psTools::multiLangField($descrip, $all_langs);
                        if (!isset($dto['metadescrip']))
                            $row['meta_description'] = Fs2psTools::multiLangField($descrip, $all_langs, array('max_length'=>255));  // TODO estos params no estaban bien: , true, 200)
                    }
                    
                    if (isset($dto['longdescrip']))
                        $row['description'] = Fs2psTools::multiLangField($dto['longdescrip'], $all_langs);
                        
                        if (isset($dto['metatitle']))
                            $row['meta_title'] = Fs2psTools::multiLangField($dto['metatitle'], $all_langs);
                            
                            if (isset($dto['metadescrip']))
                                $row['meta_description'] = Fs2psTools::multiLangField($dto['metadescrip'], $all_langs);
                                
                                if (isset($dto['metakeys']))
                                    $row['meta_keywords'] = Fs2psTools::multiLangField($dto['metakeys'], $all_langs);
                                    
                }
                
                if (isset($dto['minqty'])) $row['minimal_quantity'] = $dto['minqty'];
                
                foreach ($this->dims as $dim) {
                    if (isset($dto[$dim]))
                        $row[$dim] = $dto[$dim];
                }
                
                if (isset($dto['ean']))
                    $row['ean13'] = $dto['ean'];
                    
                    if (isset($dto['manufacturer']))
                    {
                        if (!$exists || !$this->noover_manufacturer)
                            $row['id_manufacturer'] = $this->manufacturers_upd->matcher->rowIdFromDtoId($dto['manufacturer']);
                    }
                    
                    if (isset($dto['categories']))
                    {
                        if (!$exists || !$this->noover_categories)
                            $row['_categories'] =$this->categories_upd->refsToIds($dto['categories']);
                    }
                    
                    if (isset($dto['suppliers']))
                    {
                        if (!$exists || !$this->noover_suppliers)
                            $row['_suppliers'] = $this->suppliers_upd->refsToIds($dto['suppliers']);
                    }
                    
                    
                    if (isset($dto['features']))
                    {
                        $features = array();
                        foreach ($dto['features'] as $ref => $vals)
                        {
                            $feature = array('id' => $this->features_upd->matcher->rowIdFromDtoId($ref));
                            if (is_array($vals)) {
                                $feature['value'] = $vals['value'];
                                $feature['custom'] = isset($vals['custom'])? ($vals['custom']? 1 : 0) : 1;
                            } else {
                                $feature['value'] = $vals;
                                $feature['custom'] = 1;
                            }
                            $features[] = $feature;
                        }
                        $row['_features'] = $features;
                    }
                    
                    if (isset($dto['combinations']))
                    {
                        $row['_combinations'] = $dto['combinations'];
                    }
                    
                    return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        $id = parent::insertOrUpdate($row, $exists, $oldRowId);
        
        StockAvailable::setQuantity($id, null, $row['_quantity']);
        
        $this->insertOrUpdatePrice($row, $id);
        
        if (!empty($row['_categories']))
        {
            $product = new Product($id);
            $old_categories = $product->getCategories();
            $categories = $this->categories_upd->keepNotManagedIds($row['_categories'], $old_categories);
            $product->updateCategories($categories);
            
            #if (!in_array($product->id_category_default, $categories))
            $product->id_category_default = $row['_categories'][0];
            
            $product->update();
        }
        else if (!$this->noover_categories && !$exists)
        {
            $msg = 'WARN: El producto "'.$row['_dto_id'].'" no se asoció con ninguna categoría';
            $this->task->log($msg);
        }
        
        if (!empty($row['_suppliers']))
        {
            $product = new Product($id);
            $product->deleteFromSupplier();
            foreach ($row['_suppliers'] as $id_supplier)
                $product->addSupplierReference($id_supplier, 0, null, $row['wholesale_price']);
        }
        
        if (isset($row['_features']))
        {
            $this->deleteFeatures($id);
            $this->addFeatures($id, $row['_features']);
        }
        
        $index_search = $this->task->cfg->get('PRODUCTS_UPDATE_SEARCH_INDEX');
        if ($index_search && ($index_search=='all' || ($index_search=='onlynews' and !$exists))) {
            ObjectModel::updateMultishopTable('Product', array('indexed' => 0), 'product_shop.id_product='.$id);
            Search::indexation(false, $id);
        }
        
        return $id;
    }
    
    protected function onInsertedOrUpdated($dto, $row, $inserted)
    {
        parent::onInsertedOrUpdated($dto, $row, $inserted);
        if (isset($row['_combinations'])) {
            $combinations_upd = $this->task->getUpdater('combinations');
            if (empty($row['_combinations'])) {
                $combinations_upd->removeNotPresentCombis($this->matcher->rowId($row), $row['_combinations'], true);
            } else {
                $combinations_upd->process($row['_combinations']);
            }
        }
    }
    
}


class Fs2psSizeColourCombinationUpdater extends Fs2psSKUUpdater
{
    
    protected $products_upd;
    protected $sizes_upd;
    protected $colours_upd;
    
    public static function createMatcher($task, $name) {
        // USE_CREF indica que la combinación lleva su propia referencia.
        // En este caso no se cumple que $dto_id_fields == array('ref', 'size', 'colour'), pero
        // la descarga de pedidos debería funcionar correctamente porque se basa en el mismo matcher.
        $dto_id_fields = $task->cfg->get('USE_CREF', False)? array('cref') : array('ref', 'size', 'colour');
        
        $pm = $task->cfg->get('COMBINATIONS_MATCHER');
        $matcher_cls = $pm=='basic'? 'Fs2psDto2RowMatcher' : ($pm=='pref'? 'Fs2psDto2RowPRefMatcher' : 'Fs2psDto2RowRefMatcher');
        return new $matcher_cls($name, 'product_attribute', 'id_product_attribute', $dto_id_fields, ($pm=='basic' || $pm=='pref'));
    } 
    
    public function __construct($task, $name)
    {
        $matcher = self::createMatcher($task, $name);
        parent::__construct($task, $name, 'product_attribute', $matcher->dto_id_fields, $matcher);
        
        $this->object_model_cls = 'Fs2psCombination';
        $this->products_upd = $task->getUpdater('products');
        $this->sizes_upd = $task->getUpdater('sizes');
        $this->colours_upd = $task->getUpdater('colours');
    }
    
    protected function getSizeColDesc($dto)
    {
        $attr_types = array('size', 'colour');
        $sizecoldesc = 'REF: '.$dto['ref'].' - ';
        foreach ($attr_types as $attr_type)
        {
            if (!empty($dto[$attr_type]))
                $sizecoldesc = $sizecoldesc.$dto[$attr_type].'/';
        }
        $sizecoldesc = rtrim($sizecoldesc, "/");
        return $sizecoldesc;
    }
    
    protected function getIdsAttributeFromDto($dto)
    {
        $ids_attribute = array();
        $attr_types = array('size', 'colour');
        foreach ($attr_types as $attr_type)
        {
            if (isset($dto[$attr_type]) && (!empty($dto[$attr_type]) || $dto[$attr_type]==='0'))
            {
                $upd = $attr_type.'s_upd';
                $attr_id = (int)$this->$upd->matcher->rowIdFromDtoId($dto[$attr_type]);
                if ($attr_id) $ids_attribute[] = $attr_id;
                else throw new Fs2psException('Referencia a \''.$dto[$attr_type].'\' no valida: '.$attr_id);
            }
        }
        return $ids_attribute;
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $id_product =$this->products_upd->matcher->rowIdFromDtoId($dto['ref']);
        if (empty($id_product)) {
            throw new Fs2psException(
                'No existe el producto referenciado por la combinación: '.$this->getSizeColDesc($dto)
            );
        }
            
        $ids_attribute = $this->getIdsAttributeFromDto($dto);
        if (empty($ids_attribute)) {
            throw new Fs2psException(
                'La combinación no enlaza con ningún/a talla/color existente: '.$this->getSizeColDesc($dto)
            );
        }
            
        $row = array(
            'id_product' => $this->products_upd->matcher->rowIdFromDtoId($dto['ref']),
            'reference' => $this->matcher->referenceFromDto($dto),
            
            // ¿Problems with Prestashop 1.5? Moved to onGroupUpdated
            //'default_on' => $idx == 0 ? 1 : 0,
            
            // Extra info to be used in insertOrUpdate
            '_ids_attribute' => $ids_attribute,
            '_quantity' => isset($dto['stock']) ? $dto['stock'] : 0,
            '_tax_rate' => isset($dto['iva']) ? $dto['iva'] : 0,
        );
        
        if (isset($dto['ean'])) $row['ean13'] = $dto['ean'];
        if (isset($dto['minqty'])) $row['minimal_quantity'] = $dto['minqty'];
        if (isset($dto['iva'])) $row['_tax_rate'] = $dto['iva'];
        if (isset($dto['weight'])) $row['weight'] = $dto['weight'];
        
        // Prices. Multiple price rates or simple price.
        if (empty($dto['prices'])) {
            $row['price'] = isset($dto['price']) ? $dto['price'] : 0;
        } else {
            $row['_prices'] = $dto['prices'];
            $row['price'] = $dto['prices'][0]['price'];
        }
        
        return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        $id = parent::insertOrUpdate($row, $exists, $oldRowId);
        $id_product = $row['id_product'];
        
        $combination = new Combination($id);
        
        $ids_images = $combination->getWsImages();
        $combination->setAttributes($row['_ids_attribute']);
        if ($ids_images) {
            $ids_images = array_map(function($e) { return $e['id']; }, $ids_images);
            $combination->setImages($ids_images);
        }
        
        StockAvailable::setQuantity($id_product, null, 0);
        StockAvailable::setQuantity($id_product, $id, $row['_quantity']);
        
        $this->insertOrUpdatePrice($row, $id_product, $id);
        
        return $id;
    }
    
    public function removeNotPresentCombis($id_product, $combinations, $simulate_full_process=false)
    {
        
        // Remove not uploaded combinations
        $all_combis = Fs2psTools::dbSelect('
			SELECT pa.id_product_attribute
			FROM `@DB_product_attribute` pa
			WHERE pa.`id_product` = '.(int)$id_product.'
		');
        if (!empty($all_combis))
        {
            if ($simulate_full_process) {
                $this->resetCounters($combinations);
            }
            
            $matcher = $this->matcher;
            $updated_combis_ids = array();
            foreach ($combinations as $combi) {
                $updated_combis_ids[] = $matcher->rowId($combi);
            }
            $all_combis_ids = array();
            foreach ($all_combis as $combi) {
                $all_combis_ids[] = $matcher->rowId($combi);
            }
            $delete_combis_ids = array_diff ($all_combis_ids , $updated_combis_ids);
            foreach ($delete_combis_ids as $id_combi)
            {
                $combination = new Combination($id_combi);
                $combination->delete();
                $this->ndeleted++;
            }
            
            if ($simulate_full_process) {
                $this->logProcess();
            }
        }
        
    }
    
    protected function sameGroup($row_a, $row_b)
    {
        return $row_a['id_product'] == $row_b['id_product'];
    }
    
    protected function onGroupUpdated($group_rows)
    {
        $matcher = $this->matcher;
        
        if (!empty($group_rows))
        {
            // Set default combination if none
            // Necessary in Prestashop 1.5, but not in Prestashop 1.6
            $combi = $group_rows[0];
            $id_product = $combi['id_product'];
            $default = Fs2psTools::dbValue('
				SELECT pa.id_product_attribute
				FROM @DB_product_attribute pa
				'.Shop::addSqlAssociation('product_attribute', 'pa').'
				WHERE product_attribute_shop.default_on = 1
					AND pa.id_product = '.$id_product.'
			');
            
            $product = new Product($id_product);
            
            // Price will be combination price
            $product->price = 0;
            
            // Borramos precios específicos a nivel de producto
            $reset_customer = $this->managing_price_by_customer? null : 0;
            $reset_from_quantity = $this->managing_price_from_quantity? null : 1;
            Fs2psSpecificPrice::resetProductManagedPriceRates($id_product, 0, null, 0, $reset_customer, $reset_from_quantity); # Grupo general por defecto
            Fs2psSpecificPrice::resetProductManagedPriceRates($id_product, 0, null, null, $reset_customer, $reset_from_quantity); # Grupos gestionados
            
            // Weight will be combination height if is set
            if (isset($combi['weight'])) $product->weight = 0;
            
            if (!$default)
            {
                $default = Product::getDefaultAttribute($id_product);
                $product->setDefaultAttribute($default? $default: $matcher->rowId($combi));
            }
            
            $product->update();
                        
            /* ACTUALMENTE FACTUSOL SÓLO CONTEMPLA PRECIO DE COSTE A NIVEL DE PRODUCTO
             // Set supplier references and cost unit price
             $suppliers = Fs2psTools::dbSelect('
             SELECT distinct id_supplier FROM @DB_product_supplier WHERE id_product='.$id_product.'
             ');
             if ($suppliers)
             {
             $product->deleteFromSupplier();
             foreach ($suppliers as $supplier)
             $product->addSupplierReference($id_supplier, $supplier);
             }
             */
             
        }
        
        // Remove not uploaded combinations
        $this->removeNotPresentCombis($group_rows[0]['id_product'], $group_rows);
        
        /*
        // TODO: Eliminar precios específicos a nivel de producto?
        Fs2psTools::dbExec('update ps_specific_price sp set sp.price=-1');
        */
        
    }
    
}


class Fs2psProductCriticalUpdater extends Fs2psProductUpdater
{
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array('_dto_id' => $this->matcher->dtoId($dto));
        
        // Stock
        if (isset($dto['stock'])) $row[_quantity] = $dto['stock'];
        
        // Prices and id_tax_rules_group. Multiple price rates or simple price.
        if (!empty($dto['prices'])) {
            $row['_prices'] = $dto['prices'];
            $row['price'] = $dto['prices'][0]['price'];
        } else if (isset($dto['price'])) {
            $row['price'] = $dto['price'];
        }
        if (isset($row['price'])) {
            $rate = (string)($dto['iva'] * 100);
            if (empty($this->id_tax_rules_group_by_rate[$rate])) {
                throw new Fs2psException(
                    'No se definió en la tienda un IVA del '.$rate.'% (prod. "'.$this->matcher->dtoIdToStrFromDto($dto).'")'
                    );
            }
            $row['id_tax_rules_group'] = $this->id_tax_rules_group_by_rate[$rate];
        }
        
        // cost price
        //if (isset($dto['cost'])) $row['wholesale_price'] = $dto['cost'];
        
        if (isset($dto['combinations']))  {
            $row['_combinations'] = $dto['combinations'];
        }
        
        return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        if (!$exists) throw new Fs2psContinueException();
        if (isset($row['_quantity'])) StockAvailable::setQuantity($oldRowId, null, $row['_quantity']);
        
        
        if (isset($row['price'])) $this->insertOrUpdatePrice($row, $oldRowId);
        
        // Price will be combination price
        $product->price = 0;
        
        // Borramos precios específicos a nivel de producto
        $reset_customer = $this->managing_price_by_customer? null : 0;
        $reset_from_quantity = $this->managing_price_from_quantity? null : 1;
        Fs2psSpecificPrice::resetProductManagedPriceRates($id_product, 0, null, 0, $reset_customer, $reset_from_quantity); # Grupo general por defecto
        Fs2psSpecificPrice::resetProductManagedPriceRates($id_product, 0, null, null, $reset_customer, $reset_from_quantity); # Grupos gestionados
        
        // if (isset($row['wholesale_price'])) ...
        return $oldRowId;
    }
    
}


class Fs2psSizeColourCombinationCriticalUpdater extends Fs2psSKUUpdater
{
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array(
            'reference' => $this->matcher->referenceFromDto($dto),
            '_quantity' => isset($dto['stock']) ? $dto['stock'] : 0,
        );
        
        // Prices. Multiple price rates or simple price.
        if (empty($dto['prices'])) {
            $row['price'] = isset($dto['price']) ? $dto['price'] : 0;
        } else {
            $row['_prices'] = $dto['prices'];
            $row['price'] = $dto['prices'][0]['price'];
        }
        
        return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        $id_product = $row['id_product'];
        StockAvailable::setQuantity($id_product, null, 0);
        StockAvailable::setQuantity($id_product, $oldRowId, $row['_quantity']);
        $this->insertOrUpdatePrice($row, $id_product, $oldRowId);
        return $oldRowId;
    }
    
    protected function onGroupUpdated($group_rows)
    {
        $matcher = $this->matcher;
        
        if (!empty($group_rows))
        {
            // Set default combination if none
            // Necessary in Prestashop 1.5, but not in Prestashop 1.6
            $combi = $group_rows[0];
            $id_product = $combi['id_product'];
            $default = Fs2psTools::dbValue('
				SELECT pa.id_product_attribute
				FROM @DB_product_attribute pa
				'.Shop::addSqlAssociation('product_attribute', 'pa').'
				WHERE product_attribute_shop.default_on = 1
					AND pa.id_product = '.$id_product.'
			');
            
            $product = new Product($id_product);
            
            
            // Weight will be combination height if is set
            if (isset($combi['weight'])) $product->weight = 0;
            
            if (!$default)
            {
                $default = Product::getDefaultAttribute($id_product);
                $product->setDefaultAttribute($default? $default: $matcher->rowId($combi));
            }
            
            $product->update();
        }
        
        // Remove not uploaded combinations
        $this->removeNotPresentCombis($group_rows[0]['id_product'], $group_rows);
        
        /*
         // TODO: Eliminar precios específicos a nivel de producto?
         Fs2psTools::dbExec('update ps_specific_price sp set sp.price=-1');
         */
        
    }
    
}

class Fs2psCustomerUpdater extends Fs2psUpdater
{

    protected $address_upd;
    
    public static function createMatcher($task, $name) {
        $dto_id_fields = array('ref');
        return new Fs2psDto2RowCustomerEmailMatcher($name, 'customer', 'id_customer', $dto_id_fields, true);
    }
    
    public function __construct($task, $name)
    {
        $matcher = self::createMatcher($task, $name);
        parent::__construct($task, $name, 'customer', $matcher->dto_id_fields, $matcher);
        $this->address_upd = new Fs2psAddressUpdater($task, 'customer_addresses');
        $this->default_groups = array(
            Configuration::get('PS_UNIDENTIFIED_GROUP'),
            Configuration::get('PS_GUEST_GROUP'),
            Configuration::get('PS_CUSTOMER_GROUP')
        );
    }
    
    public function getPriceRateRowId($dto_id)
    {
        $group_upd = $this->task->getUpdater('price_rates');
        return $group_upd->matcher->rowIdFromDtoId($dto_id);
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $row = array(
            'lastname' => $dto['lastname'],
            'firstname' => $dto['firstname'],
            'email' => $dto['email'],
        );
        
        if (isset($dto['company']) && !$exists) {
            // Este campo es ambiguo: Nombre fiscal? Nombre comercial?
            // Sólo establecemos su valor si se trata de un nuevo cliente,
            // en caso contrario respetamos lo que haya y no lo machacamos.
            $row['company'] = $dto['company'];
        }
        
        if (isset($dto['group'])) {
            $group_row_id = $this->getPriceRateRowId($dto['group']);
            $row['id_default_group'] = $group_row_id;
            $row['_group'] = $group_row_id;
        }
        
        if (!$exists)
        {
            $row['passwd'] = Fs2psTools::randomPassword();
        }
        
        if ($this->enable) {
            $row['active'] = 1;
        } elseif (!$exists) {
            $row['active'] = 0;
        }
        
        return $row;
    }
    
    protected function insertOrUpdate($row, $exists, $oldRowId)
    {
        $id = parent::insertOrUpdate($row, $exists, $oldRowId);
        
        $customer = new Fs2psCustomer($id);
        
        if (!$exists) {
            // Si es nuevo, le asignamos los grupos por defecto
            $customer->addGroups($this->default_groups);
        }
        
        if (isset($row['_group'])) {
            $group_id = $row['_group'];
            $groups = Fs2psCustomer::getFs2psGroups($id);
            if (count($groups)!=1 || count($groups)==1 && $groups[0]['id_group']!=$group_id) {
                if (count($groups)>0) Fs2psCustomer::resetFs2psGroups($id);
                $customer->addGroups(array($group_id));
            }
        }
        
        return $id;
    }
    
    protected function onInsertedOrUpdated($dto, $row, $inserted)
    {
        parent::onInsertedOrUpdated($dto, $row, $inserted);
        $address_row = $dto;
        $address_row['alias'] = 'Fiscal';
        $address_row['ref'] = $dto['ref'];
        $address_row['customer'] = $dto['ref'];
        $this->address_upd->process(array($address_row));
    }
    
}

class Fs2psAddressUpdater extends Fs2psUpdater
{
    
    public static function createMatcher($task, $name) {
        $dto_id_fields = array('ref');
        return new Fs2psDto2RowAddressNifMatcher($name, 'address', 'id_address', $dto_id_fields, true, $task);
    }
    
    public function __construct($task, $name)
    {
        $matcher = self::createMatcher($task, $name);
        parent::__construct($task, $name, 'address', $matcher->dto_id_fields, $matcher);
    }
    
    public function getCustomerRowId($dto_id)
    {
        $customer_upd = $this->task->getUpdater('customers');
        return $customer_upd->matcher->rowIdFromDtoId($dto_id);
    }
    
    protected function dto2row($dto, $idx, $exists, $oldRowId)
    {
        $id_country = null;
        $id_state = null;
        if (isset($dto['country'])) {
            $id_country = Country::getByIso($dto['country']);
            if (!empty($id_country) && isset($dto['state'])) {
                $id_state = State::getIdByIso($dto['state'], $id_country);
                if (empty($id_state)) {
                    $id_state = State::getIdByIso($dto['country'].'-'.$dto['state'], $id_country);
                }
            }
        }
        
        $row = array(
            'id_customer' => $this->getCustomerRowId($dto['customer']),
            'dni' => $dto['nif'],
            
            'alias' => $dto['alias'],
            'lastname' => $dto['lastname'],
            'firstname' => $dto['firstname'],
            'postcode' => $dto['postcode'],
            'city' => $dto['city'],
            
            'phone' => $dto['phone'],
            'phone_mobile' => $dto['mobile'],
            
            'address1' => $dto['address1'],
            'address2' => isset($dto['address2'])? $dto['address2'] : '',
        );
        
        if (isset($dto['company']) && !$exists) {
            // Este campo es ambiguo: Nombre fiscal? Nombre comercial?
            // Sólo establecemos su valor si se trata de una nueva dirección,
            // en caso contrario respetamos lo que haya y no lo machacamos.
            $row['company'] = $dto['company'];
        }
        
        if (!empty($id_country)) $row['id_country'] = $id_country;
        if (!empty($id_state)) $row['id_state'] = $id_state;
        
        return $row;
    }
    
    public function logProcess() {}
    
}

